"""
Abstract camera functionality.
"""
import time
import asyncio
import warnings
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.optimize import curve_fit
from scipy.ndimage import zoom
import PIL
import io
from abc import ABC, abstractmethod

from slmsuite.hardware import _Picklable
from slmsuite.holography import analysis
from slmsuite.holography.toolbox import BLAZE_LABELS, format_shape
from slmsuite.holography.toolbox.phase import zernike
from slmsuite.misc.fitfunctions import lorentzian
from slmsuite.misc.math import INTEGER_TYPES, REAL_TYPES
from slmsuite.holography.analysis import image_centroids, image_remove_field
from slmsuite.holography.analysis.files import _gray2rgb


class Camera(_Picklable, ABC):
    """
    Abstract class for cameras.
    Comes with transformations, averaging and HDR,
    and helper functions like :meth:`.autoexpose()`.

    Attributes
    ----------
    name : str
        Camera identifier.
    shape : (int, int)
        Stores ``(height, width)`` of the camera in pixels, the same convention as
        :meth:`numpy.shape`.
    bitdepth : int
        Depth of a camera pixel well in bits.
    bitresolution : int
        Returns ``(2**bitdepth) * averaging``. The action of averaging is a sum rather
        than a mean, so the effective bitresolution increases accordingly.
    dtype : np.dtype
        Stores the type returned by :meth:`._get_image_hw()`.
        This value is cached upon initialization.
    pitch_um : (float, float) OR None
        Pixel pitch in microns.
    exposure_s : float
        Caches the last result of :meth:`.get_exposure()`. Can be used if the user wants to
        avoid the overhead of calling the method.
    exposure_bounds_s : (float, float) OR None
        Shortest and longest allowable integration in seconds.
    averaging : int OR None
        Default setting for averaging (sums repeated measurements). See :meth:`.get_image()`.
    hdr : (int, int) OR None
        Default setting for multi-exposure High Dynamic Range imaging. See :meth:`.get_image()`.
    capture_attempts : int
        If the camera returns an error or exceeds a timeout,
        try again for a total of `capture_attempts` attempts.
        This is useful for resilience against errors that happen with low probability.
        Defaults to 5.
    woi : tuple
        WOI (window of interest) in ``(x, width, y, height)`` form.

        Warning
        ~~~~~~~
        This feature is less fleshed out than most. There may be issues
        (e.g. :meth:`.get_image()` with the ``averaging`` or ``hdr`` flags).
    default_shape : tuple
        Default ``shape`` of the camera before any WOI or transform changes are made.
    transform : function
        Flip and/or rotation operator specified by the user in :meth:`__init__`.
        The user is expected to apply this transform to the matrix returned in
        :meth:`get_image()`. Note that WOI changes are applied on the camera hardware
        **before** this transformation.
    last_image : numpy.ndarray OR None
        Last captured image. Note that this is a pointer to the same data that the user
        receives (to avoid copying overhead). Thus, if the user modifies the returned data,
        then this data will be modified also.
        This may be of :attr:`dtype`, or may be a float, depending on whether :attr:`hdr` is
        used and the type of :attr:`averaging`.
        Is ``None`` if no image has ever been taken.
    """
    _pickle = [
        "name",
        "shape",
        "bitdepth",
        "bitresolution",
        "pitch_um",
        "exposure_s",
        "exposure_bounds_s",
        "averaging",
        "hdr",
        "woi",
        "default_shape",
    ]
    _pickle_data = [
        "last_image",
    ]

    @abstractmethod
    def __init__(
        self,
        resolution,
        bitdepth=8,
        pitch_um=None,
        name="camera",
        exposure_bounds_s=None,
        averaging=None,
        capture_attempts=5,
        hdr=None,
        rot="0",
        fliplr=False,
        flipud=False,
    ):
        """
        Initializes a camera.

        In addition to the other class attributes, accepts the following parameters
        to set :attr:`transform`. See :meth:`~slmsuite.holography.analysis.get_orientation_transformation()`.

        Parameters
        ----------
        resolution
            The width and height of the camera in ``(width, height)`` form.

            Important
            ~~~~~~~~~
            This is the opposite of the numpy ``(height, width)``
            convention stored in :attr:`shape`.
        bitdepth
            See :attr:`bitdepth`.
        pitch_um : (float, float) OR None
            Fill in extra information about the pixel pitch in ``(dx_um, dy_um)`` form
            to use additional calibrations.
        name : str
            Defaults to ``"camera"``.
        exposure_bounds_s : (float, float) OR None
            Exposure bounds in seconds for the camera. If ``None``, no bounds are applied.
        averaging : int or None
            Number of frames to average. Used to increase the effective bit depth of a camera by using
            pre-quantization noise (e.g. dark current, read-noise, etc.) to "dither" the pixel output
            signal. If ``None``, no averaging is performed.
        hdr : int OR (int, int) OR None OR False
            Exposure information for `Multi-exposure High Dynamic Range (HDR) imaging
            <https://en.wikipedia.org/wiki/Multi-exposure_HDR_capture>`_
        capture_attempts : int
            If the camera returns an error or exceeds a timeout,
            try again for a total of `capture_attempts` attempts.
            This is useful for resilience against errors that happen with low probability.
            Defaults to 5.
        rot : str or int
            Rotates returned image by the corresponding degrees in ``["90", "180", "270"]``
            or :meth:`numpy.rot90` code in ``[1, 2, 3]``. Defaults to no rotation.
            Used to determine :attr:`shape` and :attr:`transform`.
        fliplr : bool
            Flips returned image left right.
            Used to determine :attr:`transform`.
        flipud : bool
            Flips returned image up down.
            Used to determine :attr:`transform`.
        """
        (width, height) = format_shape(resolution)

        # Set shape, depending upon transform.
        if rot in ("90", 1, "270", 3):
            self.shape = (width, height)
            self.default_shape = (width, height)
        else:
            self.shape = (height, width)
            self.default_shape = (height, width)

        # Parse capture_attempts.
        self.capture_attempts = int(capture_attempts)
        if capture_attempts <= 0:
            raise ValueError("capture_attempts must be positive.")

        # Create image transformation.
        self.transform = analysis.get_orientation_transformation(rot, fliplr, flipud)

        # Update WOI information.
        self.woi = (0, width, 0, height)
        try:
            self.set_woi()
        except NotImplementedError:
            pass

        # Variable for storing the last capture.
        self.last_image = None

        # Remember the name.
        self.name = str(name)

        # Set exposure information.
        self.exposure_bounds_s = (
            (np.min(exposure_bounds_s), np.max(exposure_bounds_s))
            if exposure_bounds_s is not None else
            None
        )

        self.exposure_s = 1     # Default to 1s for Simulated cameras.
        self.exposure_s = self.get_exposure()

        # Set datatype variables.
        self.bitdepth = int(bitdepth)
        self.dtype = self._get_dtype()

        # Frame averaging variables.
        self.averaging = self._parse_averaging(averaging, preserve_none=True)
        self.hdr = self._parse_hdr(hdr, preserve_none=True)
        self._flush_iterations = 2  # Hidden variable: how many frames to capture for a flush.

        # Spatial dimensions.
        if pitch_um is not None and not (np.isscalar(pitch_um) and pitch_um <= 0):
            if isinstance(pitch_um, REAL_TYPES):
                pitch_um = [pitch_um, pitch_um]
            self.pitch_um = np.squeeze(pitch_um)
            if len(self.pitch_um) != 2 or np.any(self.pitch_um <= 0):
                raise ValueError("Expected positive (float, float) for pitch_um")
            self.pitch_um = np.array([float(self.pitch_um[0]), float(self.pitch_um[1])])
        else:
            self.pitch_um = None

        # Placeholder for live viewer handle.
        self.viewer = None

    @property
    def bitresolution(self):
        return (2**self.bitdepth) * (self.averaging if self.averaging is not None else 1)

    # Core methods - to be implemented by subclass.

    @abstractmethod
    def close(self):
        """
        Abstract method to close the camera and delete related objects.
        """
        raise NotImplementedError()

    def __del__(self):
        try:
            self.close()
        except:
            pass

    @staticmethod
    def info(verbose=True):
        """
        Abstract method to load information about what cameras are available.

        Parameters
        ----------
        verbose : bool
            Whether or not to print display information.

        Returns
        -------
        list
            An empty list.
        """
        if verbose:
            print(".info() NotImplemented.")
        return []

    def get_exposure(self):
        """
        Get the frame integration time in seconds.
        Used in :meth:`.autoexposure()`.

        Returns
        -------
        float
            Integration time in seconds.
        """
        self.exposure_s = self._get_exposure_hw()
        return self.exposure_s

    def set_exposure(self, exposure_s):
        """
        Set the frame integration time in seconds.
        Used in :meth:`.autoexposure()`.

        Parameters
        ----------
        exposure_s : float
            The integration time in seconds.

        Returns
        -------
        float
            The resulting integration time in seconds (pulled from :meth:`.get_exposure()`).
        """
        if self.exposure_bounds_s is not None:
            exposure_s_ = np.clip(exposure_s, *self.exposure_bounds_s)
            if exposure_s_ != exposure_s:
                warnings.warn(
                    f"Requested exposure {exposure_s} s is out of bounds "
                    f"{self.exposure_bounds_s} s. Clipping to {exposure_s_} s."
                )
                exposure_s = exposure_s_
        self._set_exposure_hw(exposure_s)
        return self.get_exposure()

    @abstractmethod
    def _get_exposure_hw(self):
        """
        Abstract method to interface with hardware and get the frame integration time in seconds.
        Subclasses must implement this.
        """
        raise NotImplementedError(f"Camera {self.name} has not implemented _get_exposure_hw")

    @abstractmethod
    def _set_exposure_hw(self, exposure_s):
        """
        Abstract method to interface with hardware and set the exposure time in seconds.
        Subclasses must implement this.

        Parameters
        ----------
        exposure_s : float
            The integration time in seconds.
        """
        raise NotImplementedError(f"Camera {self.name} has not implemented _set_exposure_hw")

    def set_woi(self, woi=None):
        """
        Abstract method to narrow the imaging region to a 'window of interest'
        for faster framerates.

        Parameters
        ----------
        woi : list, None
            See :attr:`~slmsuite.hardware.cameras.camera.Camera.woi`.
            If ``None``, defaults to largest possible.

        Returns
        ----------
        woi : list
            :attr:`~slmsuite.hardware.cameras.camera.Camera.woi`.
        """
        raise NotImplementedError()

    def flush(self, timeout_s=1):
        """
        Cycle the image buffer such that all new :meth:`.get_image()` calls yield fresh frames.
        Without this feature, optimizations could be working on outdated information.

        Defaults to calling :meth:`.get_image()` twice, though cameras can implement
        hardware-specific alternatives.

        Parameters
        ----------
        timeout_s : float
            The time in seconds to wait for each frame.
            The frame exposure time  is **added** to this timeout
            such that there is always enough time to expose.
        """
        for _ in range(self._flush_iterations):
            self._get_image_hw_tolerant(timeout_s=timeout_s+self.exposure_s)

    @abstractmethod
    def _get_image_hw(self, timeout_s):
        """
        Abstract method to capture camera images.

        Parameters
        ----------
        timeout_s : float
            The time in seconds to wait for the frame to be fetched.
            The frame exposure time  is **NOT added** to this timeout
            such that there is always enough time to expose.

        Returns
        -------
        numpy.ndarray
            Array of shape :attr:`~slmsuite.hardware.cameras.camera.Camera.shape`.
        """
        raise NotImplementedError(f"Camera {self.name} has not implemented _get_image_hw")

    def _get_out(self, image_count, out=None):
        # Preallocate memory if necessary
        out_shape = (int(image_count), self.default_shape[0], self.default_shape[1])
        if out is None:
            out = np.empty(out_shape, dtype=self.dtype)
        else:
            if out.shape != out_shape:
                raise ValueError(f"Expected out to be of shape {out_shape}. Found {out.shape}.")
            if out.dtype != self.dtype:
                raise ValueError(f"Expected out to be of type {self.dtype}. Found {out.dtype}.")
            out = np.array(out, copy=False, dtype=self.dtype)

        return out

    def _get_images_hw(self, image_count, timeout_s, out=None):
        """
        Abstract method to capture a series of image_count images using camera-specific
        batch acquisition features.

        Parameters
        ----------
        image_count : int
            Number of frames to batch collect.
        timeout_s : float
            The time in seconds to wait for **each** frame to be fetched.
            The frame exposure time  is **added** to this timeout
            such that there is always enough time to expose.
        out : None OR numpy.ndarray
            Preallocated memory for in-place operations, if applicable.

        Returns
        -------
        numpy.ndarray
            Array of shape (image_count, :attr:`~slmsuite.hardware.cameras.camera.Camera.shape`).
        """
        # Preallocate memory if necessary
        out = self._get_out(image_count, out)

        for i in range(image_count):
            out[i, :, :] = self._get_image_hw_tolerant(
                timeout_s=timeout_s+self.exposure_s
            )

        return out

    # Capture methods one level of abstraction above _get_image_hw().

    def _get_image_hw_tolerant(self, *args, **kwargs):
        err = None
        failures = 0

        for _ in range(self.capture_attempts):
            try:
                img =  self._get_image_hw(*args, **kwargs)

                if failures > 0:
                    warnings.warn(f"'{self.name}' _get_image_hw() failed {failures} times before succeeding.")

                return img
            except Exception as e:
                failures += 1
                err = e

        warnings.warn(f"'{self.name}' _get_image_hw() failed {failures} times before quitting.")

        raise err

    def _get_images_hw_tolerant(self, *args, **kwargs):
        e = None
        failures = 0

        for _ in range(self.capture_attempts):
            try:
                imgs = self._get_images_hw(*args, **kwargs)

                if failures > 0:
                    warnings.warn(f"'{self.name}' _get_images_hw() failed {failures} times before succeeding.")

                return imgs
            except Exception as e:
                failures += 1
                err = e

        warnings.warn(f"'{self.name}' _get_images_hw() failed {failures} times before quitting.")

        raise err

    def _get_dtype(self, get_image_function=None):
        """
        Captures a frame from the camera to make sure the datatype conforms with
        the expected bitdepth.
        """
        if get_image_function is None:
            get_image_function = self._get_image_hw_tolerant

        try:
            self.dtype = np.dtype(
                np.array(
                    get_image_function()
                ).dtype
            )   # Future: check if cameras change dtype after init.
        except:
            if self.bitdepth <= 0:
                raise ValueError("Non-positive bitdepth does not make sense.")
            elif self.bitdepth <= 8:
                self.dtype = np.dtype(np.uint8)
            elif self.bitdepth <= 16:
                self.dtype = np.dtype(np.uint16)
            elif self.bitdepth <= 32:
                self.dtype = np.dtype(np.uint32)
            elif self.bitdepth <= 64:
                self.dtype = np.dtype(np.uint64)
            else:
                self.dtype = np.dtype(float)

        try:
            # Determine the bitdepth of the datatype.
            if self.dtype.kind == "i" or self.dtype.kind == "u":
                dtype_bitdepth = self.dtype(0).nbytes * 8
                if self.dtype.kind == "i":
                    dtype_bitdepth -= 1
            elif self.dtype.kind == "f":
                dtype_bitdepth = np.inf

            # Warn the user if something is wrong.
            if dtype_bitdepth < self.bitdepth:
                raise warnings.warn(
                    f"Camera '{self.name}' bitdepth of {self.bitdepth} does not conform "
                    f"with the image type {self.dtype} with {self.dtype.itemsize} bytes."
                )
        except:     # The above sometimes fails for non-numpy datatypes.
            pass

        return self.dtype

    def _parse_averaging(self, averaging=None, preserve_none=False):
        """
        Helper function to get a valid averaging.
        """
        if averaging is None:
            if preserve_none:
                return None
            if not hasattr(self, "averaging") or self.averaging is None:
                averaging = 1
            else:
                averaging = self.averaging
        elif averaging is False:
            averaging = 1
        averaging = int(averaging)

        if averaging <= 0:
            raise ValueError("Cannot have negative averaging.")

        return averaging

    def _parse_hdr(self, exposures=None, preserve_none=False):
        """
        Helper function to get a valid hdr parameters.
        """
        # Parse inputs
        if exposures is None:
            if preserve_none:
                return None
            if not hasattr(self, "hdr") or self.hdr is None:
                (exposures, exposure_power) = (1, 0)
            else:
                (exposures, exposure_power) = self._parse_hdr(self.hdr)
        elif exposures is False:
            exposures = 1
            exposure_power = 0
        elif np.isscalar(exposures):
            exposure_power = 2
        else:
            (exposures, exposure_power) = exposures

        # Force int so we have a chance of exposure aligning with camera clock.
        return (int(exposures), int(exposure_power))

    def _get_averaging_dtype(self, averaging=None):
        """Returns the appropriate image datatype for ``averaging`` levels of averaging."""
        if averaging is None:
            if self.averaging is None:
                raise ValueError("Averaging is not enabled for this camera. Set the .averaging attribute.")
            else:
                averaging = self.averaging
        averaging = int(averaging)

        if averaging <= 0:
            raise ValueError("Cannot have negative averaging.")

        # Get dtype instance from type if needed
        dtype = np.dtype(self.dtype) if not hasattr(self.dtype, 'kind') else self.dtype

        # Switch based on image type
        if dtype.kind == "i" or dtype.kind == "u":
            dtype_bitdepth = 8*dtype.type(0).nbytes

            # Remove depth for signed integer.
            if dtype.kind == "i":
                dtype_bitdepth -= 1

            extra_bits = int(np.rint(np.log2(averaging)))

            if self.bitdepth + extra_bits <= dtype_bitdepth:
                # If we can sustain the averaging with the current type, continue.
                return self.dtype
            else:
                # Otherwise, force floating point.
                return float
        elif dtype.kind == "f":
            # Return floating point.
            return self.dtype
        else:
            raise ValueError(f"Datatype {self.dtype} does not make sense as a camera return.")

    def get_image(self, timeout_s=1, transform=True, hdr=None, averaging=None):
        """
        Capture, process, and return images from a camera.

        Tip
        ~~~
        This function includes two advanced capture options:

        -   `Multi-exposure High Dynamic Range (HDR) imaging
            <https://en.wikipedia.org/wiki/Multi-exposure_HDR_capture>`_
            and
        -   Software frame averaging (integrating).

        These methods can aid the user in capturing more precise data, beyond the
        default raw (and bitdepth-limited) output of the camera.

        Parameters
        ----------
        timeout_s : float
            The time in seconds to wait for the frame to be fetched.
            The frame exposure time  is **added** to this timeout
            such that there is always enough time to expose.
        transform : bool
            Whether or not to transform the output image according to
            :attr:`~slmsuite.hardware.cameras.camera.Camera.transform`.
            Defaults to ``True``.
        hdr : int OR (int, int) OR None OR False
            Exposure information for `Multi-exposure High Dynamic Range (HDR) imaging
            <https://en.wikipedia.org/wiki/Multi-exposure_HDR_capture>`_
            If ``None``, the value of :attr:`hdr` is used.
            If ``False``, HDR is not used no matter the state of :attr:`hdr`.

            See Also
            ~~~~~~~~
            :meth:`.get_image_hdr()` for more information.

        averaging : int OR None OR False
            If ``int``, the number of frames to average over.
            If ``None``, the value of :attr:`averaging` is used.
            If ``False``, averaging is not used no matter the state of :attr:`averaging`.

            Tip
            ~~~
            The datatype is promoted to float if necessary but otherwise tries to stick
            with the default datatype.
            For instance, a camera that returns a 12-bit image as a 16-bit type has four
            more bits to use for averaging, i.e. :math:`2^4 = 16` possible averages without
            risk of overflow.
            Requesting more than 16 averages would cause the return type to be promoted
            to ``float``.

            Important
            ~~~~~~~~~
            This feature sums many measurements together (does not mean),
            thereby averaging without floating point operations.
            This is done such that integer datatypes (useful for memory compactness) can still be returned,
            whereas a general mean would need to be floating point.

        Returns
        -------
        numpy.ndarray of int OR float
            Array of shape :attr:`~slmsuite.hardware.cameras.camera.Camera.shape`.
        """
        # Parse acquisition options.
        averaging = self._parse_averaging(averaging)
        (exposures, exposure_power) = self._parse_hdr(hdr)

        # Switch based on what imaging case we're in.
        if exposures > 1:       # Average many images with increasing exposure.
            return self.get_image_hdr(
                (exposures, exposure_power),
                timeout_s=timeout_s,
                transform=transform,
                averaging=averaging,
            )
        elif averaging > 1:     # Average many images.
            averaging_dtype = self._get_averaging_dtype(averaging)

            try:
                # Using the camera-specific batch method if available
                imgs = self._get_images_hw(
                    averaging, timeout_s=timeout_s+self.exposure_s
                ).astype(averaging_dtype)

                # Cast as the proper type so we can sum.
                img = np.sum(imgs, axis=0)
            except NotImplementedError:
                # Brute-force collection as a backup
                img = np.zeros(self.default_shape, dtype=averaging_dtype)

                for _ in range(averaging):
                    img += self._get_image_hw_tolerant(
                        timeout_s=timeout_s+self.exposure_s
                    ).astype(averaging_dtype)
        else:                   # Normal image
            img = self._get_image_hw_tolerant(
                timeout_s=timeout_s+self.exposure_s
            )

        # self.transform implements the flipping and rotating keywords passed to the
        # superclass constructor.
        if transform:
            img = self.transform(img)

        # Store the result locally.
        self.last_image = img

        # Push to viewer if active.
        if self.viewer is not None:
            if averaging > 1:
                self.viewer.render(img / averaging)
            else:
                self.viewer.render(img)

        return img

    def get_images(self, image_count, timeout_s=1, out=None, transform=True, flush=False):
        """
        Grab ``image_count`` images in succession.

        Important
        ~~~~~~~~~
        This method **does not** support averaging or HDR features.
        Rather, it just returns a series of raw images.

        Parameters
        ----------
        image_count : int
            Number of images to grab.
        timeout_s : float
            The time in seconds to wait **for each** frame to be fetched.
            The frame exposure time  is **added** to this timeout
            such that there is always enough time to expose.
        out : None OR numpy.ndarray
            If not ``None``, output data in this memory. Useful to avoid excessive allocation.
        transform : bool
            Whether or not to transform the output image according to
            :attr:`~slmsuite.hardware.cameras.camera.Camera.transform`.
            Defaults to ``True``.
        flush : bool
            Whether to flush before grabbing.

        Returns
        -------
        numpy.ndarray
            Array of shape ``(image_count, height, width)``.
        """
        # Flush if desired.
        if flush:
            self.flush()

        # Grab images (no transformation)
        imgs = self._get_images_hw(
            image_count,
            timeout_s=timeout_s+self.exposure_s,
            out=out
        )

        # Transform if desired. Future: make more efficient.
        if transform:
            imgs_ = np.empty(
                (int(image_count), self.shape[0], self.shape[1]),
                dtype=self.dtype
            )
            for i in range(image_count):
                imgs_[i, :, :] = self.transform(imgs[i])

            imgs = imgs_

        # Store the result locally.
        self.last_image = imgs[-1]

        # Push to viewer if active.
        if self.viewer is not None:
            self.viewer.render(imgs[-1])

        return imgs

    def get_image_hdr(self, exposures=None, return_raw=False, **kwargs):
        r"""
        Often, the necessities of precision applications exceed the bitdepth of a
        camera. One way to recover High Dynamic Range (HDR) imaging is to use
        `multiple exposures <https://en.wikipedia.org/wiki/Multi-exposure_HDR_capture>`_
        each with increasing exposure time. Then, these images can be stitched together
        as floating-point data, omitting data which is under- or over- exposed.

        Tip
        ~~~
        This feature can be accessed in :meth:`.get_image()`
        using :attr:`hdr` or the ``hdr=`` flag.
        This function is exposed here also to reveal the raw data using ``return_raw=``
        and to draw attention to this useful feature.

        Caution
        ~~~~~~~
        Camera exposure is sometimes poorly defined. This might cause incorrect
        assumptions of the exposure.
        In general, a larger base exposure will produce more accurate results as a
        greater number of sample clock periods are rounded to for smaller relative variation.
        Future modifications to :meth:`get_image_hdr_analysis()` might improve image stitching.

        Parameters
        ----------
        exposures : int OR (int, int)
            The number of exposures to take.
            Each exposure increases in time multiplicatively from the base value
            (original :meth:`get_exposure()`) by a factor :math:`p`.
            The :math:`i\text{th}` image has exposure time :math:`\tau \times p^i`, zero-indexed.
            The default base of :math:`p = 2` leads to ``exposures`` being equivalent to
            `spots <https://en.wikipedia.org/wiki/Exposure_value>`_.
            This base can be changed to another number by instead passing a tuple, where
            the second ``int`` defines the desired base.
        return_raw : bool
            If ``True``, returns the raw data (stack of images with count ``exposures``)
            instead of the processed data. The data can be processed using :meth:`get_image_hdr_analysis`
        **kwargs
            Passed to :meth:`.get_image()`.

        Returns
        -------
        numpy.ndarray of float
            Array of shape :attr:`~slmsuite.hardware.cameras.camera.Camera.shape`.

            Important
            ~~~~~~~~~
            The scale of the returned image is the same as the original exposure.
        """
        (exposures, exposure_power) = self._parse_hdr(exposures)
        overexposure_threshold = self.bitresolution / 2
        if self.averaging is not None:
            overexposure_threshold *= self.averaging

        # Make empty data and grab the original exposure time.
        original_exposure = self.get_exposure()
        imgs = np.zeros((exposures, self.shape[0], self.shape[1]), self.dtype)
        exposure_times = np.zeros((exposures,), dtype=float)

        for i in range(exposures):
            # FUTURE: record the set exposures and use these to do better analysis.
            exposure_times[i] = self.set_exposure(int(exposure_power ** i) * original_exposure)
            self.flush()    # Sometimes, cameras return bad frames after exposure change.
            imgs[i, :, :] = self.get_image(hdr=False, **kwargs)

            # Terminate the loop if our image is entirely overexposed.
            if np.all(imgs[i, :, :] > overexposure_threshold):
                continue

        # Reset exposure.
        self.set_exposure(original_exposure)

        if return_raw:
            return imgs, exposure_times
        else:
            img = self.get_image_hdr_analysis(
                imgs,
                overexposure_threshold=overexposure_threshold,
                exposure_power=exposure_times,
            )
            if np.max(img) >= self.bitresolution:
                warnings.warn("HDR image is overexposed.")
            # Store the result locally.
            self.last_image = img
            return img

    @staticmethod
    def get_image_hdr_analysis(imgs, overexposure_threshold=None, exposure_power=2):
        r"""
        Analyzes raw data for High Dynamic Range (HDR) imaging
        `multiple exposures <https://en.wikipedia.org/wiki/Multi-exposure_HDR_capture>`_
        each with increasing exposure time.

        Parameters
        ----------
        imgs : array_like
            Stack of images with increasing exposure.
        overexposure_threshold : float OR None
            For each image (except the first), data is thrown out if values are above
            this threshold. If ``None``, the threshold defaults to half the maximum.
        exposure_power : int or list of float
            Each exposure increases in time multiplicatively from the base value
            (original :meth:`get_exposure()`) by this factor :math:`p`. The :math:`i\text{th}` image has
            exposure time :math:`\tau \times p^i`, zero-indexed.
            The default value of ``2`` leads to ``exposures`` being equivalent to
            `spots <https://en.wikipedia.org/wiki/Exposure_value>`_.

        Returns
        -------
        numpy.ndarray of float
            Array of shape :attr:`~slmsuite.hardware.cameras.camera.Camera.shape`.

            Important
            ~~~~~~~~~
            The scale of the returned image is the same as the original exposure.
        """
        # Parse arguments
        if np.isscalar(exposure_power):
            exposure_power = float(int(exposure_power))
            exposure_times = np.power(exposure_power, np.arange(imgs.shape[0]))
        else:
            exposure_times = np.array(exposure_power)
            if np.all(exposure_times <= 0):
                raise ValueError("exposure_times cannot all be non-positive.")
            exposure_times = exposure_times / np.min(exposure_times[exposure_times > 0])

        if overexposure_threshold is None:
            # Default to half exposure.
            overexposure_threshold = np.max(imgs) / 2

        img = None

        for i in range(imgs.shape[0]):
            img_current = imgs[i, :, :].astype(float)

            if i == 0:
                img = img_current
            elif exposure_times[i] > 0:
                # Overwrite data when greater precision is available.
                mask = img_current < overexposure_threshold
                img[mask] = img_current[mask] / exposure_times[i]

        return img

    # Self-test method to test everything above.

    def test(self):
        """
        Tests the core hardware methods of :class:`Camera`.
        Validates that the camera is connected correctly and all hardware
        features are supported.
        """
        print(f"Testing camera: {self.name}")

        # Test 1: Exposure methods (required)
        print("  Testing exposure methods...")
        current_exposure = self.get_exposure()
        assert isinstance(current_exposure, REAL_TYPES)
        assert current_exposure > 0

        # Test setting exposure
        new_exposure = current_exposure * 1.5
        set_exposure = self.set_exposure(new_exposure)
        assert isinstance(set_exposure, REAL_TYPES)

        # Restore original exposure
        self.set_exposure(current_exposure)

        # Test 2: Capture methods (requires concrete implementation)
        print("  Testing capture methods...")
        orig_averaging = self.averaging
        orig_hdr = self.hdr

        self.averaging = None
        self.hdr = None
        self.last_image = None

        # Test basic image capture
        print("    Testing get_image...")
        image = self.get_image(timeout_s=2)
        assert isinstance(image, np.ndarray)
        assert image.shape == self.shape
        assert image.dtype == self.dtype or np.issubdtype(image.dtype, np.floating)

        # Test that last_image is updated
        assert self.last_image is not None
        if np.issubdtype(image.dtype, np.floating) or np.issubdtype(self.last_image.dtype, np.floating):
            assert np.allclose(self.last_image, image)
        else:
            assert np.array_equal(self.last_image, image)

        # Test get_image with transform=False
        image_no_transform = self.get_image(transform=False, timeout_s=2)
        assert isinstance(image_no_transform, np.ndarray)

        # Test averaging
        print("    Testing averaging...")
        image_avg = self.get_image(averaging=2, timeout_s=5)
        assert isinstance(image_avg, np.ndarray)
        assert image_avg.shape == self.shape

        # Test get_images
        print("    Testing get_images...")
        image_count = 3
        images = self.get_images(image_count, timeout_s=2)
        assert isinstance(images, np.ndarray)
        assert images.shape == (image_count, self.shape[0], self.shape[1])
        assert images.dtype == self.dtype

        # Test get_images with preallocated output
        out = np.empty((image_count, self.shape[0], self.shape[1]), dtype=self.dtype)
        images_out = self.get_images(image_count, timeout_s=2, out=out)
        assert images_out.shape == out.shape
        assert images_out.dtype == out.dtype

        # Test flush method
        print("    Testing flush...")
        self.flush(timeout_s=2)

        # Test HDR imaging
        try:
            print("    Testing get_image_hdr...")
            hdr_image = self.get_image_hdr(exposures=2, return_raw=False, timeout_s=3)
            assert isinstance(hdr_image, np.ndarray)
            assert hdr_image.shape == self.shape
            assert np.issubdtype(hdr_image.dtype, np.floating)

            # Test HDR with return_raw=True
            hdr_raw, exposure_times = self.get_image_hdr(exposures=2, return_raw=True, timeout_s=3)
            assert isinstance(hdr_raw, np.ndarray)
            assert isinstance(exposure_times, np.ndarray)
            assert hdr_raw.shape[0] == 2
            assert hdr_raw.shape[1:] == self.shape
            assert len(exposure_times) == 2

        except Exception as e:
            print(f"      HDR testing skipped: {e}")

        # Test 3: Set WOI (if implemented)
        print("  Testing set_woi...")
        try:
            orig_woi = getattr(self, 'woi', None)
            self.set_woi()
            if orig_woi is not None:
                self.set_woi(orig_woi)
        except NotImplementedError:
            print("    set_woi not implemented - skipping")

        # Test 4: Info method
        print("  Testing info...")
        result = self.info(verbose=False)
        assert isinstance(result, list)

        print("Camera test completed successfully!")

        return True

    # Display methods.

    def plot(self, image=None, limits=None, title="Image", ax=None, cbar=True):
        """
        Plots the provided image.

        Parameters
        ----------
        image : ndarray OR None OR bool
            Image to be plotted.
            If ``None``, grabs an image from the camera.
            If ``False``, uses the :attr:`.last_image` attribute.
        limits : None OR float OR [[float, float], [float, float]]
            Scales the limits by a given factor or uses the passed limits directly.
        title : str
            Title the axis.
        ax : matplotlib.pyplot.axis OR None
            Axis to plot upon.
        cbar : bool
            Also plot a colorbar.

        Returns
        -------
        matplotlib.pyplot.axis
            Axis of the plotted image.
        """
        if image is None:
            self.flush()
            image = self.get_image()
        if image is False:
            image = self.last_image
        image = np.array(image, copy=(False if np.__version__[0] == '1' else None))

        if len(plt.get_fignums()) > 0:
            fig = plt.gcf()
        else:
            fig = plt.figure(figsize=(20,8))

        if ax is not None:
            plt.sca(ax)

        im = plt.imshow(image)
        ax = plt.gca()

        if cbar:
            cax = make_axes_locatable(ax).append_axes("right", size="2%", pad=0.05)
            fig.colorbar(im, cax=cax, orientation="vertical")

        # ax.invert_yaxis()
        ax.set_title(title)

        if limits is not None and limits != 1:
            if np.isscalar(limits):
                axlim = [ax.get_xlim(), ax.get_ylim()]

                centers = np.mean(axlim, axis=1)
                deltas = np.squeeze(np.diff(axlim, axis=1)) * limits / 2

                limits = np.vstack((centers - deltas, centers + deltas)).T
            elif np.shape(limits) == (2,2):
                pass
            else:
                raise ValueError(f"limits format {limits} not recognized; provide a scalar or limits.")

            ax.set_xlim(limits[0])
            ax.set_ylim(limits[1])

        if image.shape == self.shape:
            ax.set_xlabel(BLAZE_LABELS["ij"][0])
            ax.set_ylabel(BLAZE_LABELS["ij"][1])

        plt.sca(ax)

        return ax

    def live(self, activate=None, widgets=True, backend="ipython", **kwargs):
        """
        Creates and displays an IPython camera viewer.
        This viewer displays the result of :meth:`get_image()`
        or the last image of :meth:`get_images()` **whenever these methods are called**.
        Averaging and HDR are displayed with the same color scaling as without.

        If ``True`` is passed to the ``widgets`` argument, this viewer is accompanied by
        a series of `IPython widgets
        <https://ipywidgets.readthedocs.io/en/latest/examples/Widget%20List.html>`_
        in the form of slides and buttons
        for controlling the color scale, colormap, viewer scale, and live viewing.
        By toggling the ``Live`` widget button,
        this viewer can be used as a realtime camera monitor within the jupyter notebook.
        Note that any user-execution will block the monitoring loop.
        Regardless, any image polling during the blocked period will still update the viewer,
        which provides useful active feedback for what is happening during the execution.

        This limitation is imposed by the
        python Global Interpreter Lock (GIL) which restricts operation to a single thread,
        especially operation connecting to a diverse set of camera and SLM hardware.
        We use :mod:`asyncio` to allow the realtime monitoring loop to be
        interrupted by user-execution (e.g. running a cell in jupyter),
        blocking until the execution is finished.

        Running multiple viewers at once might not play nicely right now.

        Parameters
        ----------
        activate : bool OR None
            If ``True``, creates a live viewer in the current cell,
            destroying any other attached viewer.
            If ``False``, destroys  any other attached viewer.
            If ``None``, toggles the live viewer, destroying any attached viewer or
            creating one in the current cell if none is attached. Defaults to ``None``.
        widgets : bool
            If ``True``, also displays sliders and controls used to hone the display properties.
        backend : str
            Placeholder option for different types of viewers.
            The default is ``"ipython"``.
        **kwargs
            Options passed to the :class:`_CameraViewer` to customize the default settings.
            These features will be made less hidden in the future.
            Most things are customizable via these keywords. For instance, the user can pass
            a custom list of colormaps to appear in the widget dropdown as ``cmap_options=``.
        """
        if backend != "ipython":
            raise ValueError(
                f"'{backend}' not recognized; "
                "'ipython' is currently the only supported .live() backend."
            )

        try:
            from ipywidgets import Image
            from IPython.display import display
        except ImportError:
            raise ImportError("jupyter must be installed to use .live().")

        if (self.viewer is None and activate is None) or activate:
            if self.viewer is not None:
                self.viewer.close()

            self.viewer = _CameraViewer(
                self,
                widgets,
                backend,
                **kwargs
            )
        elif self.viewer is not None and (activate is None or not activate):
            self.viewer.close()
            self.viewer = None

    # Other helper methods.

    def autoexposure(
        self,
        set_fraction=0.5,
        tol=0.05,
        exposure_bounds_s=None,
        window=None,
        timeout_s=5,
        verbose=True,
    ):
        """
        Sets the exposure of the camera such that the maximum value is at ``set_fraction``
        of the dynamic range. Useful for mitigating over- or under- exposure.

        Parameters
        --------
        set_fraction : float
            Fraction of camera dynamic range to use as a target image maximum.
        tol : float
            Fractional tolerance for exposure adjustment.
        exposure_bounds_s : (float, float) OR None
            Shortest and longest allowable integration in seconds. If ``None``, defaults to
            :attr:`exposure_bounds_s`. If this attribute was not set (or not available on
            a particular camera), then ``None`` instead defaults to unbounded.
        window : array_like or None
            See :attr:`~slmsuite.hardware.cameras.camera.Camera.woi`.
            If ``None``, the full camera frame will be used.
        timeout_s : float
            Stop attempting adjusting exposure after ``timeout_s`` seconds.
        verbose : bool
            Whether to print exposure updates.

        Returns
        --------
        float
            Resulting exposure in seconds.
        """
        # Parse exposure_bounds_s
        if exposure_bounds_s is None:
            if self.exposure_bounds_s is None:
                exposure_bounds_s = (0, np.inf)
            else:
                exposure_bounds_s = self.exposure_bounds_s

        # Parse window
        if window is None:
            wxi = 0
            wxf = self.shape[1]
            wyi = 0
            wyf = self.shape[0]
        else:
            wxi = int(window[0] - window[1] / 2)
            wxf = int(window[0] + window[1] / 2)
            wyi = int(window[2] - window[3] / 2)
            wyf = int(window[2] + window[3] / 2)

        # Initialize loop
        set_val = 0.5 * self.bitresolution
        exp = self.get_exposure()
        self.flush()
        img = self.get_image()
        im_max = np.amax(img[wyi:wyf, wxi:wxf])

        # Calculate the error as a percent of the camera's bitresolution
        err = np.abs(im_max - set_val) / self.bitresolution
        t = time.perf_counter()

        # Loop until timeout expires or we meet tolerance
        while err > tol and time.perf_counter() - t < timeout_s:
            # Clip exposure steps to 0.5x -> 2x
            exp = exp / np.amax([0.5, np.amin([(im_max / set_val), 2])])
            exp_desired = exp
            exp = np.clip(exp, exposure_bounds_s[0], exposure_bounds_s[1])
            if exp_desired != exp:
                raise RuntimeError(f"autoexposure has railed (exposure: {exp_desired}, bounds: {exposure_bounds_s}).")

            self.set_exposure(exp)
            self.flush()
            img = self.get_image()

            im_max = np.amax(img[wyi:wyf, wxi:wxf])
            err = np.abs(im_max - set_val) / self.bitresolution

            if verbose:
                print(
                    f"Autoexposure: exposure = {exp:<.2e} s, "
                    f"image_max = {im_max}/{self.bitresolution-(self.averaging if self.averaging is not None else 1)}, ",
                )

        # The loop targets 50% of resolution. Now set the final exposure if different.
        if set_fraction != 0.5:
            exp = exp * (2 * set_fraction)
            self.set_exposure(exp)

        return exp

    @staticmethod
    def _autofocus_metric(img, plot=False):
        """See :meth:`.autofocus()` ``metric=``"""
        dft = np.fft.fftshift(np.fft.fft2(img.astype(float)))
        dft_amp = np.abs(dft)
        dft_norm = dft_amp / np.amax(dft_amp)
        fom = np.sum(dft_norm)

        if plot:
            _, axs = plt.subplots(1, 2)

            axs[0].imshow(img)
            axs[0].set_title("Image")
            axs[0].set_xticks([])
            axs[0].set_yticks([])

            axs[1].imshow(dft_norm)
            axs[1].set_title(f"FFT\nFoM$ = \\int\\int $|FFT|$ / $max|FFT|$ = {fom}$")
            axs[1].set_xticks([])
            axs[1].set_yticks([])

            plt.show()

        return fom

    def autofocus(self, set_z, get_z=0, range_z=2, metric=None, plot=False, verbose=False):
        """
        Finds optimal focus when scanning over some variable ``z``.
        This ``z`` often takes the form of a vertical stage to position a sample precisely
        at the plane of imaging of a lens or objective.
        The default ``metric`` is based on the Fourier contrast of the image,
        and works particularly well when combined with a projected spot array hologram.

        Parameters
        ----------
        set_z : function OR SLM
            Sets the position of the focusing stage to a given ``float``.
            If an SLM is passed, adds a lens phase to the SLM to focus the existing
            pattern. In this case, the units of ``z`` are in Zernike defocus terms
            (wavefronts). The optimal defocus is added to the wavefront calibration
            (``source["phase"]``) of the SLM.
        get_z : function OR float
            Gets the current position of the focusing stage. Should return a ``float``.
            Can also pass a ``float`` representing the center of the search range.
        range_z : array_like OR float OR None
            ``z`` values to sweep over during search relative to the base position ``get_z``.
            If a single ``float`` is passed, sweeps from ``-range_z`` to ``+range_z``
            with 11 steps.
        metric : function OR None
            Function which evaluates the focus quality of an image.
            Should take in an image and return a scalar figure-of-merit (FoM).
            Defaults to :meth:`Camera._autofocus_metric`, which approximates the
            sharpness of the images (implemented as the Fourier contrast of the image,
            the sum of the normalized Fourier amplitudes). The sharper the image, the
            higher the FoM.
        plot : bool
            Whether to provide illustrative plots.

        Returns
        -------
        float
            Optimal ``z`` value found.
        """
        # Parse set_z
        if hasattr(set_z, 'set_phase'):
            # SLM passed; create lens phase setter.
            slm = set_z
            base_phase = slm.phase.copy()
            base_correction = slm.source.get('phase', np.zeros_like(base_phase))
            base_phase -= base_correction

            def slm_set_z(z_val):
                slm.source['phase'] = (
                    base_correction +
                    zernike(slm, index=4, weight=z_val, use_mask=False)
                )
                slm.set_phase(
                    base_phase,
                    settle=True
                )

            set_z = slm_set_z

        if not callable(set_z):
            raise ValueError("set_z must be a function or SLM.")

        # Parse get_z
        z_base = get_z
        if callable(get_z):
            z_base = get_z()

        # Parse range_z
        z_list = range_z
        if np.isscalar(range_z):
            z_list = np.linspace(-range_z, range_z, 11, endpoint=True)
        z_list += z_base
        z_list = sorted(z_list)

        # Parse metric
        if metric is None:
            metric = Camera._autofocus_metric

        # Setup for the sweep
        imlist = []
        counts = np.full_like(z_list, np.nan)

        for i, z in enumerate(z_list):
            try:
                if verbose:
                    print(f"Moving to z = {z:<.2f}...          ", end="\r")
                set_z(z)

                # Take image and evaluate metric.
                self.flush()
                img = self.get_image()
                imlist.append(np.copy(img))
                counts[i] = metric(img)
            except:
                pass

        # Handle the case where everything failed.
        if np.all(np.isnan(counts)):
            try:
                set_z(z_base)
            except:
                pass
            raise RuntimeError("Autofocus failed; no valid images captured.")

        # Otherwise, fit a Lorentzian to the data to find the optimum.
        I_max_count = np.nanargmax(counts)

        dz = np.mean(np.diff(z_list))
        popt0 = np.array(
            [z_list[I_max_count], np.max(counts) - np.min(counts), np.min(counts), (z_list[-1]-z_list[0])]
        )
        bounds = np.array(
            [
                [z_list[0], 0, 0, dz],
                [z_list[-1], (np.max(counts) - np.min(counts))*2, np.max(counts), np.inf]
            ]
        )

        # try:
        if True:
            popt, _ = curve_fit(
                lorentzian,
                z_list,
                counts,
                bounds=bounds,
                ftol=1e-5,
                p0=popt0,
            )
            z_opt = popt[0]
            c_opt = popt[1] + popt[2]
        # except BaseException:
        #     if verbose:
        #         print("Autofocus fit failed, using maximum fom as optimum...")
        #     z_opt = z_list[I_max_count]
        #     c_opt = counts[I_max_count]

        # Goto the optimal position
        if verbose:
            print("Moving to optimized value, z = " + str(z_opt))
        set_z(z_opt)

        # Show result if desired
        if plot:
            plt.plot(z_list, counts, label="Data")
            plt.xlabel(r"$z$")
            plt.ylabel("Figure of Merit")
            plt.title("Autofocus Sweep")
            plt.scatter(z_opt, c_opt, label="Result")

            lfit = None
            try:
                lfit = lorentzian(z_list, *popt)
            except BaseException:
                lfit = None
            if lfit is not None:
                plt.plot(z_list, lfit, label="Fit")
            plt.legend()
            plt.show()

        return z_opt


class _CameraViewer:
    """
    Hidden class for live camera viewing enabled by ipython widgets.
    """
    def __init__(
            self,
            cam,
            widgets,
            backend="ipython",
            live=False,
            min=None,
            max=None,
            log=False,
            cmap=True,
            scale=1,
            border=None,
            cmap_options=[
                "default", "gray", "Blues", "turbo",
                'viridis', 'plasma', 'inferno', 'magma', 'cividis'
            ],
            crosshair=False,
            centroid=False,
        ):
        self.cam = cam
        self.backend = backend

        # Parse range.
        if min is None:
            min = 0
        if max is None:
            max = cam.bitresolution-1
        range = [min, max]
        range = [np.min(range), np.max(range)]

        if cmap is True: cmap = "default"
        if cmap is False: cmap = "grayscale"

        # Parse scale
        scale = 2 ** np.round(np.log2(scale))

        self.state = {
            "backend" : backend,
            "live" : live,
            "range" : range,
            "log" : bool(log),
            "cmap" : cmap,
            "scale" : scale,
            "border" : border,
            "cmap_options" : cmap_options,
            "center_crosshair" : crosshair,
            "centroid_crosshair" : centroid,
        }

        self.task = None
        self.widgets = {}
        if widgets: self.init_widgets()
        self.init_image()

    def parse(self, img=None):
        if img is not None:
            self.prev_img = img
        if self.prev_img is None:
            return  # Nothing to render.

        # Downscaling can happen before intensive operations.
        if self.state["scale"] < 1:
            img = zoom(
                self.prev_img,
                [self.state["scale"], self.state["scale"]] + ([1] if len(self.prev_img.shape) == 3 else []),
                order=1
            )
        else:
            img = np.copy(self.prev_img)

        if self.state["centroid_crosshair"]:
            img_median_subtract = image_remove_field([img], deviations=None)
            cx, cy = np.rint(
                (
                    np.squeeze(image_centroids(img_median_subtract)) + np.flip(img.shape) / 2
                ) * (self.state["scale"] if self.state["scale"] > 1 else 1)
            ).astype(int)

        # Scale intensity of image
        r = np.array(self.state["range"]).astype(img.dtype)
        img = np.clip(img, r[0], r[1])
        img -= r[0]
        d = r[1] - r[0]

        if self.state["log"]:
            # clip to avoid log(0)
            img = (np.log10(np.clip(img, 1, np.inf)) / np.log10(d+1))

        # Make image color
        rgb = _gray2rgb(
            img,
            cmap=self.state["cmap"],
            lut=d,
            normalize=False,
            border=self.state["border"]
        )

        # Upscaling can happen after intensive operations.
        if self.state["scale"] > 1:
            rgb = zoom(rgb, (1, self.state["scale"], self.state["scale"], 1), order=0)

        # Add crosshair at the median-subtracted centroid (center of mass) position.
        if self.state["centroid_crosshair"]:
            rgb[:, :, cx, :3] = 255 - rgb[:, :, cx, :3]
            rgb[:, cy, :, :3] = 255 - rgb[:, cy, :, :3]

        # Finally, add crosshair in the center.
        if self.state["center_crosshair"]:
            rgb[:, :, int(rgb.shape[2]/2), :3] = 127 - rgb[:, :, int(rgb.shape[2]/2), :3]
            rgb[:, int(rgb.shape[1]/2), :, :3] = 127 - rgb[:, int(rgb.shape[1]/2), :, :3]

        buff = io.BytesIO()
        rgb = PIL.Image.fromarray(rgb[0])
        rgb.save(buff, format="png")

        return buff.getvalue()

    def render(self, img=None):
        try:
            self.image.value = self.parse(img)
        except Exception as e:
            with self.widgets["output"]:
                print(str(e))

    def update(self, event):
        with self.widgets["output"]:
            self.widgets["output"].clear_output(wait=True)
        for key in ["range", "log", "cmap", "scale", "live", "center_crosshair", "centroid_crosshair"]:
            self.state[key] = self.widgets[key].value

        self.render()

    def live(self, event=None):
        state = self.state["live"] = self.widgets["live"].value

        loop = asyncio.get_running_loop()

        if self.task is not None:
            try:
                self.task.cancel()
            except:
                pass

        if not state:
            self.task = None
        else:
            self.task = loop.create_task(self.live_loop())

    async def live_loop(self):
        while self.state["live"]:
            self.cam.get_image()
            await asyncio.sleep(0.01)

    def on_click(self, event):
        coord = np.array([event["x"], event["y"]])
        with self.widgets["output"]:
            self.widgets["output"].clear_output(wait=True)
            print(np.round(coord / self.state["scale"]).astype(int))

    def autorange(self, event):
        if self.prev_img is not None:
            range = [np.min(self.prev_img), np.max(self.prev_img)]
            self.state["range"] = self.widgets["range"].value = range

        self.render()

    def init_image(self):
        from ipywidgets import Image
        from IPython.display import display

        self.image = Image(value=self.parse(self.cam.get_image()), format="png")
        self.image.on_click = self.on_click
        display(self.image)

    def init_widgets(self):
        from ipywidgets import HTML, IntRangeSlider, ToggleButton, Button, Checkbox, Dropdown, FloatLogSlider, Output, Layout

        item_layout = Layout(width="auto")
        range_layout = Layout(width="70%")

        self.widgets = {
            "name" : HTML(
                value=f"<b>{self.cam.name}</b>",
                description="Viewing",
                tooltip="Name of the camera.",
                layout=item_layout,
            ),
            "live" : ToggleButton(
                value=self.state["live"],
                description="Live",
                tooltip="Toggle an asyncio loop to poll images from the camera.",
                layout=item_layout,
            ),
            "range" : IntRangeSlider(
                value=self.state["range"],
                min=0,
                max=self.cam.bitresolution-1,
                step=1,
                description="Range",
                tooltip="Color scale of the plot.",
                layout=range_layout,
            ),
            "autorange" : Button(
                description="AutoRange",
                tooltip="Scale the plot to the minimum and maximum of the current image.",
                layout=item_layout,
            ),
            "log" : Checkbox(
                value=self.state["log"],
                description="Logarithmic",
                tooltip="Toggle logarithmic scaling of the current plot.",
                layout=item_layout,
            ),
            "center_crosshair" : Checkbox(
                value=self.state["center_crosshair"],
                description="Center Crosshair",
                tooltip="Toggle a crosshair centered on the image.",
                layout=item_layout,
            ),
            "centroid_crosshair" : Checkbox(
                value=self.state["centroid_crosshair"],
                description="Centroid Crosshair",
                tooltip="Toggle a crosshair at the median-subtracted centroid (center of mass) of the image.",
                layout=item_layout,
            ),
            "cmap" : Dropdown(
                options=self.state["cmap_options"],
                value=self.state["cmap"],
                description="Colormap",
                tooltip="Choose the colormap to use for display.",
                layout=item_layout,
            ),
            "scale" : FloatLogSlider(
                value=self.state["scale"],
                base=2,
                min=-3, # 12.5%
                max=3,  # 800%
                step=1,
                description="Scale",
                tooltip="Scale the image by powers of two.",
                layout=item_layout,
            ),
            "output": Output()
        }

        for k, w in self.widgets.items():
            if k == "autorange":
                w.on_click(self.autorange)
            elif k == "live":
                w.observe(self.live, "value")
            else:
                w.observe(self.update, "value")

        from ipywidgets import HBox, VBox
        from IPython.display import display

        # self.widgets["layout"] = VBox([
        #     HBox([
        #         self.widgets["name"],
        #         self.widgets["cmap"],
        #         self.widgets["log"],
        #         self.widgets["scale"],
        #     ]),
        #     HBox([
        #         self.widgets["range"],
        #         self.widgets["autorange"],
        #     ]),
        #     self.widgets["output"],
        # ])

        box_layout1 = Layout(
            display="flex",
            flex_flow="auto",
            align_items="stretch",
            width="70%"
        )
        box_layout2 = Layout(
            display="flex",
            flex_flow="auto",
            align_items="stretch",
            width="30%"
        )

        self.widgets["layout"] = HBox([
            VBox(
                [
                    HBox([
                        self.widgets["name"],
                    ]),
                    HBox([
                        self.widgets["cmap"],
                        self.widgets["log"],
                        self.widgets["center_crosshair"],
                        self.widgets["centroid_crosshair"],
                    ]),
                    HBox([
                        self.widgets["range"],
                    ]),
                    self.widgets["output"],
                ],
                layout=box_layout1,
            ),
            VBox(
                [
                    self.widgets["live"],
                    self.widgets["scale"],
                    self.widgets["autorange"],
                ],
                layout=box_layout2,
            )
        ])

        display(self.widgets["layout"])

    def close(self):
        try:
            self.task.cancel()
            self.task = None
        except:
            pass

        for w in self.widgets.values():
            w.close()
        self.image.close()

        del self



from typing import Any
from slmsuite.holography.algorithms._header import *
from slmsuite.holography.algorithms._hologram import Hologram
from slmsuite.holography.algorithms._spots import SpotHologram, CompressedSpotHologram
from slmsuite.holography.algorithms._feedback import FeedbackHologram


class MultiplaneHologram(Hologram):
    """
    Holography combining multiple objectives, potentially across planes of focus or color.
    Other :class:`Hologram` subclasses are restricted to either optimizing a hologram
    within a fixed basis of spots or
    within the grid of a discrete Fourier transform at a fixed plane of focus.
    This :class:`MultiplaneHologram` acts as a metaclass to optimize many individual
    holograms simultaneously---over many planes or pointsets---producing a composite
    phase pattern.

    Note
    ~~~~
    Though the infrastructure to make this trivial is not yet in place,
    the idea of a 'plane' extends to planes of color. That is, this class
    :class:`MultiplaneHologram` could also be used to optimize a multicolor hologram and
    account for how the farfield of each color scales with wavelength.

    Tip
    ~~~
    Calls to :meth:`.optimize()` which update :attr:`flags` also update the flags of any
    child hologram.

    Attributes
    ----------
    holograms : list of :class:`Hologram`
        List of sub-holograms to optimize simultaneously.
    weights : list of float
        Weight for each hologram. This allows the user to redistribute power between
        holograms. Keep in mind that each hologram will normalize itself, so differences
        in intensity between target patterns cannot be relied upon.
    """
    def __init__(self, holograms, weights=None):
        """
        Initializes a 'meta' hologram consisting of several sub-holograms optimizing at
        the same time.

        Parameters
        ----------
        holograms : list of :class:`Hologram`
            List of ``N`` sub-holograms to optimize simultaneously.
        weights : array_like of float OR None
            List of ``N`` floats.
            If ``None``, defaults to even power.
        """
        self.holograms = holograms

        # Check that all holograms are actually holograms and not MultiplaneHolograms.
        for h in self.holograms:
            if "MultiplaneHologram" in str(type(h)):
                raise ValueError("Multiplane hologram recursion is not supported.")
            if not "Hologram" in str(type(h)):
                raise ValueError(f"Multiplane hologram must be provided child holograms, not {type(h)}")

        # Construct the parent hologram with empty goals but complete context.
        super().__init__(
            target=holograms[0].slm_shape,      # This hologram has a fake target.
            amp=holograms[0].amp,
            phase=holograms[0].phase,
            slm_shape=holograms[0].slm_shape,
            dtype=holograms[0].dtype,
        )
        self.target = None

        # Force all the child holograms to point to the same data.
        for h in self.holograms:
            h.amp = self.amp
            h.phase = self.phase

        # Parse weights
        if weights is None:
            weights = np.ones(len(self), dtype=self.dtype)

        self.weights = np.array(weights, copy=(False if np.__version__[0] == '1' else None), dtype=self.dtype)
        self.weights /= Hologram._norm(self.weights, xp=np)

    def __len__(self):
        return len(self.holograms)

    @staticmethod
    def get_multiplane_defocus_blur(
        cameraslm,
        targets,
        target_depths,
        return_depths=None,
        sharp_focus=True,
    ):
        """
        From a stack of target (power) images at ``target_depths``, generate a stack
        of images at ``return_depths``, accounting for defocus blur.
        Power is summed as if all depths were transparent; i.e. objects do no block
        objects further behind.
        This is a partial farfield implementation of
        `realistic defocus blur <https://doi.org/10.48550/arXiv.2205.07030>`_.

        Warning
        -------
        This feature seems to lead to less stable holography without, perhaps, some
        additional optimizations.

        Parameters
        ----------
        cameraslm : ~slmsuite.hardware.cameraslms.FourierSLM
            Hardware to implement blur for. Calibrations are necessary to determine how
            much to blur.

            Tip
            ---
            Right now, the blurring is Gaussian and analytic, but in the future, the
            measurement of the point spread function should be used.
        targets : array_like
            Stack of images of shape ``(image_count, h, w)``.
        target_depths : list of float
            Depths in ``"kxy"`` focal power units corresponding to the ``targets``.
        return_depths : list of float
            Depths to return images at.
            If ``None``, use ``target_depths``.
        sharp_focus : bool
            If ``False``, depths at focal planes are blurred by the point spread radius
            of a focussed spot.
            If ``True``, all the blurring is reduced by the focused point spread radius,
            keeping images that are in focus sharp.
        """
        # Parse return_depths.
        if return_depths is None:
            return_depths = target_depths

        # Parse targets.
        if len(np.shape(targets)) != 3:
            raise ValueError("Expected 3D stack of 2D images.")

        (image_count, h, w) = np.shape(targets)

        # Check target_depths.
        if image_count != len(target_depths):
            raise ValueError("There should be the same number of images as target_depths.")

        # Make the return data and gather useful parameters.
        canvas = np.zeros((len(return_depths), h, w))

        # Gather f_eff in cam_pix/rad.
        if cameraslm.cam.pitch_um is None:
            raise ValueError(
                "Camera pitch_um is necessary to calculate defocus blur. "
                "Otherwise, we have no reference for the scale of a wavelength."
            )

        f_eff = np.sqrt(np.abs(np.linalg.det(cameraslm.calibrations["fourier"]["M"])))
        w0_kxy = cameraslm.slm.get_spot_radius_kxy()
        w0_pix = f_eff * w0_kxy
        w0_um = w0_pix * np.mean(cameraslm.cam.pitch_um)

        zr = np.pi * w0_um * w0_um / cameraslm.slm.wav_um     # (what if n != 1?)

        for j, z2 in enumerate(return_depths):
            for i, z1 in enumerate(target_depths):
                dz = (z1 - z2) * (f_eff * f_eff)

                blur = w0_pix * (np.sqrt(1 + (dz / zr) ** 2) - (1 if sharp_focus else 0))
                blur = 2 * int(blur) + 1

                canvas[j, :, :] += cv2.GaussianBlur(targets[i], (blur, blur), 0)

        return canvas

    # Overload user functions with meta functionality.

    def _update_flags(self, method, verbose, feedback, stat_groups, **kwargs):
        # First update the parent flags.
        super()._update_flags(method, verbose, feedback, stat_groups, **kwargs)

        # Then update each of the child flags.
        for h in self.holograms:
            h.flags.update(self.flags)

    def _update_weights(self, *args, **kwargs):
        for h in self.holograms: h._update_weights(*args, **kwargs)

    def _gs_farfield_routines(self, *args, **kwargs):
        for h in self.holograms: h._gs_farfield_routines(*args, **kwargs)

    def _get_target_moments_knm_norm(self):
        # Get the data from the child holograms.
        centers = []
        stds = []
        for h in self.holograms:
            center, std = h._get_target_moments_knm_norm()
            centers.append(center)
            stds.append(std)

        # Weight the centers.
        centers = np.vstack(centers)
        center = np.nansum(np.square(self.weights).reshape(-1, 1) * centers, axis=0)

        # With the center, now weight the stds. We're doing an analytic integration of
        # x^2 over rectangles corresponding to the center \pm sqrt(3) * std of each hologram.
        stds = np.vstack(stds)

        c = centers - center.reshape(1, 2)
        l = c - stds * np.sqrt(3)
        r = c + stds * np.sqrt(3)

        integral_normalized = (r * r * r - l * l * l) / (2 * stds * np.sqrt(3)) / 3
        std = np.sqrt(np.nansum(np.square(self.weights).reshape(-1, 1) * integral_normalized, axis=0))

        return center, std

    def reset(self, reset_phase=True, reset_flags=False):
        # Resetting the phase of the parent resets the phase of the children because
        # phase is shared.
        super().reset(reset_phase, reset_flags)

        # Reset the other child variables.
        for h in self.holograms: h.reset(reset_phase=False, reset_flags=reset_flags)

    def reset_weights(self):
        for h in self.holograms: h.reset_weights()

    def plot_farfield(self, *args, **kwargs):
        for h in self.holograms: h.plot_farfield(*args, **kwargs)

    # def plot_nearfield(self, *args, **kwargs):
    #     for h in self.holograms: h.plot_nearfield(*args, **kwargs)

    def plot_stats(self, *args, **kwargs):
        for h in self.holograms: h.plot_stats(*args, **kwargs)

    def _update_stats(self, stat_groups=[]):
        # FUTURE: make meta stat group.
        for h in self.holograms: h._update_stats(stat_groups)

    def set_target(self, *args, **kwargs):
        raise RuntimeError(
            "Do not use MultiplaneHologram.set_target(). "
            "Instead, update the targets of the children holograms directly."
        )

    # Multiplane hacks to get meta optimization to work.

    def _cg_loss(self, phase_torch):
        """Sum the losses of all the child holograms."""
        loss = self.holograms[0]._cg_loss(phase_torch)

        for h in self.holograms[1:]:
            loss += h._cg_loss(phase_torch)

        return loss

    def _nearfield2farfield(self):
        """Have all the holograms populate their own farfield variables."""
        for h in self.holograms:
            h._nearfield2farfield()
            h.iter = self.iter

    def _farfield2nearfield(self):
        """Sum all the complex nearfields together for the meta nearfield."""
        self.nearfield.fill(0)

        for h, w in zip(self.holograms, self.weights):
            h._farfield2nearfield(extract=False)    # Avoid individually extracting phase.

            (i0, i1, i2, i3) = toolbox.unpad(h.shape, h.slm_shape)

            # Add the complex individual nearfields to our meta nearfield.
            if h.propagation_kernel is None:
                self.nearfield += w * h.nearfield[i0:i1, i2:i3]
            else:
                # Remove the propagation kernel if necessary.
                self.nearfield += w * h.nearfield[i0:i1, i2:i3] * cp.exp(-1j * h.propagation_kernel)
            h.iter = self.iter

        # Get meta self phase.
        self._nearfield_extract()

    def _mraf_helper_routines(self):
        return [h._mraf_helper_routines() for h in self.holograms]

    def _gs_farfield_routines(self, mraf_variables):
        for h, mraf in zip(self.holograms, mraf_variables):
            h._gs_farfield_routines(mraf)

    def remove_vortices(self):
        for h in self.holograms: h.remove_vortices()



"""
Provides a TCPIP client-server interface to control remote hardware.
Interface with a :class:`~slmsuite.hardware.remote.Server` using
:class:`~slmsuite.hardware.slms.remote.RemoteSLM` and
:class:`~slmsuite.hardware.cameras.remote.RemoteCamera`.
Under the hood, this uses :mod:`json` formatting with
:mod:`zlib` compression used for any :mod:`numpy` data, including especially
megapixel phase masks and megapixel camera images.

Danger
~~~~~~
This interface will operate best in a secure and trusted local network.
Hosting a server on a public IP raises some risk of tampering
(even with an ``allowlist``, as IP addresses can be spoofed).
Communication is locked down to only the few commands that are required
by an abstract camera or SLM to reduce risk.
However, there are no protections against DDOS attack or similar and
communication is **not encrypted or authenticated**.

In the future, if there is interest, SSH authentication could be added, and the scope of
commands offered could be expanded beyond the required abstract commands
to also include whatever hardware-specific functionality is
implemented. Moreover, CameraSLMs could be hosted on a server such that system
calibrations could live in the cloud.

Example
~~~~~~~
The following is a simple example of a setup communicating between two
threads---for example, two Jupyter notebooks---on the same computer (via ``localhost:5025``).

The server hosts a simulated SLM and camera:

.. highlight:: python
.. code-block:: python

    # Server notebook
    from slmsuite.hardware.slms.simulated import SimulatedSLM
    from slmsuite.hardware.cameras.simulated import SimulatedCamera
    from slmsuite.hardware.remote import Server

    slm = SimulatedSLM((1600, 1200), pitch_um=(8,8), name="remote_slm")
    cam = SimulatedCamera(slm, (1440, 1100), pitch_um=(4,4), gain=50, name="remote_camera")

    server = Server(
        hardware=[slm, cam],
        port=5025,
    )

The client connects to this hardware:

.. highlight:: python
.. code-block:: python

    # Client notebook
    from slmsuite.hardware.slms.remote import RemoteSLM
    from slmsuite.hardware.cameras.remote import RemoteCamera

    slm = RemoteSLM(
        name="remote_slm"
        host="localhost",
        port=5025,
    )
    cam = RemoteCamera(
        name="remote_camera"
        host="localhost",
        port=5025,
    )

    slm.set_phase(None)

    cam.get_image()
    cam.plot()
"""
import numpy as np
import socket, sys, json, time
import warnings
import urllib.parse as urllib
from datetime import date, datetime, timedelta
import traceback
from typing import Any, List, Tuple, Dict
import zlib
import base64

from slmsuite.hardware import _Picklable
from slmsuite import __version__

DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 5025             # Commonly used for instrument control.
DEFAULT_TIMEOUT = 5
SERVER_WAIT_TIMEOUT = 0.5

_delim = "\n"

# Common functions for encoding and decoding data.
def _recurse_decompress(msg):
    """
    Recursively decompresses the result of json serialization.
    """
    if isinstance(msg, dict):
        if "__zlib__" in msg and len(msg) == 3:
            return np.frombuffer(
                zlib.decompress(
                    base64.b64decode(msg["__zlib__"])
                ),
                dtype=np.dtype(msg["__dtype__"])
            ).reshape(msg["__shape__"])
        elif "__dtype__" in msg and len(msg) == 1:
            return np.dtype(msg["__dtype__"])
        else:
            for k in msg:
                msg[k] = _recurse_decompress(msg[k])
    elif isinstance(msg, list):
        for i, m in enumerate(msg):
            msg[i] = _recurse_decompress(m)

    return msg

# https://codetinkering.com/numpy-encoder-json/
class _NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.bool_):
            return bool(obj)
        if isinstance(obj, np.floating): #, np.complexfloating
            return float(obj)
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.ndarray):
            return {
                "__zlib__" : base64.b64encode(
                    zlib.compress(obj.tobytes())
                ).decode(),
                "__shape__" : obj.shape,
                "__dtype__" : str(obj.dtype)
            }
        if isinstance(obj, np.string_):
            return str(obj)
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        if isinstance(obj, timedelta):
            return str(obj)
        if isinstance(obj, np.dtype):
            return {"__dtype__" : str(obj)}
        return super(_NpEncoder, self).default(obj)

# Common function to receive data from a socket.
def _recv(sock, timeout):
    recv_buffer = 4096 * 64
    buffer = ""
    t = time.time()

    # Pull data into the buffer until we hit timeout or deliminator.
    while time.time() - t < timeout:
        data = sock.recv(recv_buffer).decode()

        buffer += data
        if data[-1] == _delim:
            msg = json.loads(urllib.unquote_plus(buffer[0:-len(_delim)]))

            msg = _recurse_decompress(msg)

            return msg

    # Failed timeout returns empty.
    return False, f"Timeout: {len(buffer)} bytes received."

# Server which hosts slmsuite hardware.
class Server:
    """
    Server for handling client commands and interfacing with hardware.
    """

    def __init__(
            self,
            hardware: List[object],
            port: int = DEFAULT_PORT,
            timeout: float = SERVER_WAIT_TIMEOUT,
            allowlist: List[str] = None,
        ):
        """
        Initializes a server to host slmsuite hardware: cameras and SLMs.
        Interface with this server using
        :class:`~slmsuite.hardware.slms.remote.RemoteSLM` and
        :class:`~slmsuite.hardware.cameras.remote.RemoteCamera`.

        :param hardware:
            List of hardware objects to serve.
        :param port:
            Port number to serve on. Defaults to ``5025``
            (commonly used for instrument control).
        :param timeout:
            Timeout in seconds for the server to wait for a client.
        :param allowlist:
            List of IP addresses to allow to connect. Defaults to ``None`` (allow all).
            Keep in mind that IP addresses can be spoofed, so this ``allowlist``
            provides only modest security.
        """
        # Parse hardware.
        for hw in hardware:
            if not hasattr(hw, "name"):
                raise ValueError(f"Hardware {hw} must have a 'name' attribute.")
            if self.identify_hardware(hw) is None:
                raise ValueError(f"Hardware {hw.name} ({hw}) must be either a camera or an SLM.")

        names = [hw.name for hw in hardware]
        if len(set(names)) != len(names):
            raise ValueError(f"Hardware names must be unique. Found {names}.")

        self.hardware = {
            hw.name : hw
            for hw in hardware
        }
        self.kind = {
            hw.name : self.identify_hardware(hw)
            for hw in hardware
        }

        # Server information.
        if not (1024 <= port <= 65535):
            raise ValueError(f"Invalid port number: {port}. Use a port between 1024 and 65535.")
        self.port = port
        self.timeout = timeout

        # Only allow clients in the allowlist to connect.
        self.allowlist = allowlist

        # Only allowed commands for SLMs and Cameras, alongside "ping".
        self.allowcommands = [
            "pickle",
            "flush",
            "_set_phase_hw",
            "_set_exposure_hw",
            "_get_exposure_hw",
            "_get_image_hw",
            "_get_images_hw",
        ]

    @staticmethod
    def identify_hardware(hw: object) -> str:
        if hasattr(hw, "_get_image_hw"):
            return "camera"
        elif hasattr(hw, "_set_phase_hw"):
            return "slm"
        else:
            return None

    def listen(self, verbose: bool = True):
        """
        Blocking command to listen for client commands and process them once they are
        given.

        :param verbose:
            Whether to print feedback that the server is online alongside a log of
            client actions.
        """
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.settimeout(self.timeout)
        sock.bind(("", self.port))
        sock.listen(5)

        i = 0

        if verbose:
            print(f"Hosting on port {self.port} with hardware {list(self.hardware.keys())}")

        try:
            while True:
                i += 1
                try:
                    # Wait for connection with fancy print.
                    if verbose:
                        print("Waiting for connection" + ("." * (1 + (i % 3))) + "     ", end="\r")

                    # This blocks for self.timeout unless a client connects.
                    connection, client_addr = sock.accept()

                    # Cleanup the print.
                    if verbose:
                        print("                              ", end="\r")

                    # Check if the client is allowed to connect.
                    if (self.allowlist is not None) and (client_addr[0] not in self.allowlist):
                        if verbose:
                            stamp = str(datetime.now())
                            print(f"{stamp} Rejected connection from {client_addr}, not in allowlist {self.allowlist}.")
                        result = False, f"Client {client_addr} not in allowlist."
                    else:
                        # Receive, handle, and reply to message.
                        message = _recv(connection, self.timeout)
                        result = self._handle(message, client_addr, verbose)

                    reply = (urllib.quote_plus(json.dumps(result, cls=_NpEncoder)) + _delim).encode()
                    print(f"replied with {len(reply)} bytes.")

                    connection.sendall(reply)
                    connection.close()
                except IOError as e:
                    # This is a timeout error. Just continue.
                    pass
                except Exception as e:
                    # Pass to the outer try for all other errors.
                    raise e
        except KeyboardInterrupt:
            # Standard way to kill the thread.
            if verbose:
                print("Closing server! Goodbye!")
            try:
                connection.close()
            except:
                pass
            sock.close()
        except Exception as e:
            # There was an error in the server communication protocol. This kills the thread.
            # Note that hardware errors are handled in _handle and the loop continues.
            if verbose:
                print(traceback.format_exc())
            try:
                connection.close()
            except:
                pass
            sock.close()
            raise e

    def _handle(
        self,
        message : str,
        client_addr: str = None,
        verbose: bool = False
    ) -> Tuple[bool, Any]:
        """
        Handle a message from a client.
        """
        try:
            name = message.pop("name", None)
            command = message.pop("command", None)
            args = message.pop("args", [])
            kwargs = message.pop("kwargs", dict())

            instrument = f"{name}.{command}"

            if verbose:
                stamp = str(datetime.now())
                print(f"{stamp} {client_addr} {instrument}")

            # Initial parse of command.
            if command is None:
                return False, "No command provided."
            elif command == "ping":
                return True, self.kind

            # Make sure that the hardware exists.
            if not name in self.hardware:
                return False, f"Did not recognize hardware '{name}'. Options: {list(self.hardware.keys())}."

            if command in self.allowcommands and hasattr(self.hardware[name], command):
                attribute = getattr(self.hardware[name], command)
                if callable(attribute):
                    return True, attribute(*args, **kwargs)
                else:
                    return False, f"{instrument} is not callable."
            else:
                return False, f"{instrument} not present."
        except:
            return False, traceback.format_exc()

# Abstract client which connects to a server.
class _Client(_Picklable):
    """
    Client for interfacing with a slmsuite server.
    """

    def __init__(
            self,
            name: str,
            kind: str,
            host: str = DEFAULT_HOST,
            port: int = DEFAULT_PORT,
            timeout: float = DEFAULT_TIMEOUT,
        ):
        """
        Superclass constructor. See RemoteSLM and RemoteCamera for more information.
        """
        self.name = name
        self.host = host
        self.port = port
        self.timeout = timeout

        hardware = self._com(command="ping")

        if not self.name in hardware:
            raise ValueError(
                f"Hardware '{self.name}' is not present at {self.host}:{self.port}. Options: {hardware}."
            )
        if hardware[self.name] != kind:
            raise ValueError(
                f"Hardware '{self.name}' is not a {kind} at {self.host}:{self.port}."
            )

        try:
            t = time.perf_counter()
            pickled = self._com(
                command="pickle",
                kwargs=dict(attributes=False, metadata=True)
            )
            t = time.perf_counter() - t
        except:
            raise RuntimeError(
                f"Could not connect to '{self.name}' at {self.host}:{self.port}. Options: {hardware}."
            )

        self.latency_s = t
        self.server_attributes = pickled

        if not "__version__" in pickled:
            warnings.warn(
                f"Server did not provide version information; "
                f"cannot verify compatibility with client version {__version__}."
            )
        elif pickled["__version__"] != __version__:
            warnings.warn(
                f"Client version {__version__} does not match server version {pickled['__version__']}."
            )

    def _com(
        self,
        command: str = "ping",
        args: list = [],
        kwargs: dict = {},
    ):
        """Helper function to _com without having to put all the name/host information in."""
        return _Client.__com(self.name, self.host, self.port, self.timeout, command, args, kwargs)

    @staticmethod
    def __com(
        name: str,
        host: str = DEFAULT_HOST,
        port: int = DEFAULT_PORT,
        timeout: float = DEFAULT_TIMEOUT,
        command: str = "ping",
        args: list = [],
        kwargs: dict = {},
    ):
        """Generalized function to communicate with a server."""
        # Create a TCP/IP socket
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(timeout)
        try:
            sock.connect((host, port))
        except (TimeoutError, ConnectionRefusedError):
            raise ValueError(
                f"An slmsuite server is not active at {host}:{port}."
            )
        except Exception as e:
            raise e


        # Send the message.
        sock.sendall((
            urllib.quote_plus(
                json.dumps(
                    {
                        "name": name,
                        "command": command,
                        "args": args,
                        "kwargs": kwargs
                    },
                    cls=_NpEncoder
                )
            ) + _delim
        ).encode())


        # Wait for a reply.
        try:
            success, reply = _recv(sock, timeout)
            if success == False:
                raise RuntimeError(
                    f"Server {host}:{port} communication failed. Message:\n{reply}"
                )
        except Exception as e:
            sock.close()

            raise e

        # Always close.
        sock.close()

        return reply

    @staticmethod
    def info(
        host: str = DEFAULT_HOST,
        port: int = DEFAULT_PORT,
        timeout: float = DEFAULT_TIMEOUT,
        verbose: bool = True,
    ) -> Dict[str, str]:
        """
        Looks to see if a slmsuite server is active at the given host and port.

        :param verbose:
            Whether to print the discovered information.
        :param host:
            Which host to connect to. Defaults to ``"localhost"``.
        :param port:
            Which port to connect to. Defaults to ``5025`` (commonly used for instrument
            control).
        :param timeout:
            Timeout in seconds for the connection. Defaults to ``1.0``.
        :return:
            List of hardware at the server in ``name:kind`` pairs, where ``kind`` is
            either ``"camera"`` or ``"slm"``. Returns empty dict if no server is found.
        """
        try:
            hardware = _Client.__com(None, host, port, timeout, command="ping")
        except (TimeoutError, ConnectionRefusedError):
            raise TimeoutError(f"Did not find a server at {host}:{port}.")
        except Exception as e:
            raise e

        if verbose:
            if len(hardware) == 0:
                print(f"Server found at {host}:{port} with no hardware.")
            else:
                print(
                    f"Server found at {host}:{port} with hardware:\n    " +
                    "\n    ".join(list(hardware.keys()))
                )

        return hardware


"""
GPU-accelerated holography algorithms.

This module is currently focused on
`Gerchberg-Saxton (GS) <http://www.u.arizona.edu/~ppoon/GerchbergandSaxton1972.pdf>`_
iterative Fourier transform phase retrieval algorithms
via the :class:`~slmsuite.holography.algorithms.Hologram` class;
however, support for complex holography and other algorithms
(e.g. `gradient descent algorithms <https://doi.org/10.1364/AO.21.002758>`_)
is also planned. Additionally, so-called Weighted Gerchberg-Saxton (WGS) algorithms for hologram
generation with or without closed-loop camera feedback are supported, especially for
the `generation of optical focus arrays <https://doi.org/10.1364/OL.44.003178>`_,
a subset of general image formation. We also support `Mixed Region Amplitude Freedom (MRAF)
<https://doi.org/10.1007/s10043-018-0456-x>`_ feedback.

Tip
~~~
This module makes use of the GPU-accelerated computing library :mod:`cupy`
(`GitHub <https://docs.cupy.dev/en/stable/reference/index.html>`_).
If :mod:`cupy` is not supported, then :mod:`numpy` is used as a fallback, though
CPU alone is significantly slower. Using :mod:`cupy` is highly encouraged.
The only algorithm that has no CPU fallback is :class:`CompressedSpotHologram`
in a Zernike basis beyond 2D or 3D spots. Other cases can use the CPU (however slowly).

Note
~~~~
Internally, algorithms is split into several hidden files
to enhance clarity and reduce file length.

- ``_header.py`` : The common imports for all the files.
- ``_stats.py`` : Statistics and plotting common to all hologram classes.
- ``_hologram.py`` : The core file. Contains the actual algorithms (:class:`Hologram`).
- ``_feedback.py`` : Infrastructure for image feedback (:class:`FeedbackHologram`).
- ``_spots.py`` : Infrastructure for spot-specific holography
  (:class:`SpotHologram`, :class:`CompressedSpotHologram`).
"""
from slmsuite.holography.algorithms._header import *

from slmsuite.holography.algorithms._hologram import Hologram as _Hologram
from slmsuite.holography.algorithms._feedback import FeedbackHologram as _FeedbackHologram
from slmsuite.holography.algorithms._spots import SpotHologram as _SpotHologram
from slmsuite.holography.algorithms._spots import CompressedSpotHologram as _CompressedSpotHologram
from slmsuite.holography.algorithms._multiplane import MultiplaneHologram as _MultiplaneHologram

# Hack to get automodule to put the classes in the correct location.
class Hologram(_Hologram):
    pass

class FeedbackHologram(_FeedbackHologram):
    pass

class SpotHologram(_SpotHologram):
    pass

class CompressedSpotHologram(_CompressedSpotHologram):
    pass

class MultiplaneHologram(_MultiplaneHologram):
    pass

# Hack to get the class and attribute docs to work.
Hologram.__doc__ = _Hologram.__doc__
FeedbackHologram.__doc__ = _FeedbackHologram.__doc__
SpotHologram.__doc__ = _SpotHologram.__doc__
CompressedSpotHologram.__doc__ = _CompressedSpotHologram.__doc__
MultiplaneHologram.__doc__ = _MultiplaneHologram.__doc__


"""
**(Untested)** Hardware control for The Imaging Source cameras via :mod:`tisgrabber`.
:mod:`tisgrabber` is one of several different interfaces that The Imaging Source supports.
See
`the tisgrabber source
<https://github.com/TheImagingSource/IC-Imaging-Control-Samples/tree/master/Python/tisgrabber>`_.
This was tested at commit 7846b9e and Python 3.9 with DMK 27BUP031 camera.
The tisgrabber .dll and tisgrabber.py are needed.
Please either install tisgrabber.py or have it in your current working directory.
"""
import warnings
import ctypes
import numpy as np

from slmsuite.hardware.cameras.camera import Camera

try:
    import tisgrabber as tis
except:
    tis = None
    warnings.warn("tisgrabber not installed. Install to use ImagingSource cameras.")


# Change this DLL path if necessary
DLL_PATH = "./tisgrabber_x64.dll"

class ImagingSource(Camera):
    """
    The Imaging Source camera.

    Attributes
    ----------
    sdk : ctypes.CDLL
        Connects to the Imaging Source SDK. Shared among instances of :class:`ImagingSource`.
    cam : HGRABBER
        Object to talk with the camera. See tisgrabber.h or tisgrabber documentation for more details
    vid_format : str
        Caches the video format currently set by the user if known.
    """
    sdk = None

    @classmethod
    def init_sdk(cls):
        """
        Class method for initializing the sdk. Called when the first instance is instantiated or when the static method info is called.
        Parameters
        ----------
        cls : object
            required parameter for a class method.

        Raises
        ------
        RuntimeError
           If the library fails to initiate. See tisgrabber.h for error codes.
        """
        sdk = ctypes.cdll.LoadLibrary(DLL_PATH)
        tis.declareFunctions(sdk)

        err = sdk.IC_InitLibrary(0)
        if err != 1:
            raise Exception("DLL library failed to initiate. Perhaps check the DLL_PATH in tis_camera.py")

        cls.sdk = sdk

        return err

    @staticmethod
    def safe_call(cb, to_raise, *args, **kwargs):
        """
        Decorator method that automatically error checks the result from callback `cb`.

        Parameters
        ----------
        cb : function
            Function that is decorated with arguments `*args` and `**kwargs`.
        to_raise : bool
            Whether to raise an exception or simply print out an error.

        Returns
        -------
        err : int
            error code is returned regardless when Exception is raised. Error code information is in tisgrabber.h.
        """
        err = cb(*args, **kwargs)
        if err <= 0:
            err_str = "Error performing operation: " + cb.__name__ + " err code: " + str(err)
            if to_raise:
                raise Exception(err_str)
            else:
                print(err_str)
        return err

    def __init__(
        self,
        serial="",
        vid_format=None,
        pitch_um=None,
        verbose=True,
        **kwargs
    ):
        """
        Initialize camera and attributes.

        Parameters
        ----------
        serial : str
            This serial is used to open a camera by unique name (see tisgrabber.h).
            It is usually the model name followed by a space and the serial number.
            Use :meth:`.info()` to see detected options.
            If empty, then opens the first camera found.
        vid_format : str
            If None, no format is set and will default to whatever the camera is currently.
            See tisgrabber.h for more information. Example ``"Y800 (2592x1944)"``.
        pitch_um : (float, float) OR None
            Fill in extra information about the pixel pitch in ``(dx_um, dy_um)`` form
            to use additional calibrations.
        verbose : bool
            Whether or not to print extra information.
        **kwargs
            See :meth:`.Camera.__init__` for permissible options.
        """
        if tis is None:
            raise ImportError("tisgrabber not installed. Install to use ImagingSource cameras.")

        # Initialize the SDK if needed.
        if verbose: print("TIS Camera SDK initializing... ", end="")
        if ImagingSource.sdk is None:
            err = ImagingSource.init_sdk()
            if err != 1:
                raise Exception("Error when loading SDK: " + str(err))
        if verbose: print("success")

        # Then we load the camera from the SDK.
        if verbose: print('"{}" initializing... '.format(serial), end="")

        # cam will be the handle that represents the camera.
        self.cam = ImagingSource.sdk.IC_CreateGrabber()
        if serial == "":
            connected_devs = ImagingSource.info()
            if len(connected_devs) == 0:
                raise Exception("No cameras found")
            serial = connected_devs[0] # By default use the first camera that is found
        err = ImagingSource.sdk.IC_OpenDevByUniqueName(self.cam, tis.T(serial))
        if err != 1:
            raise Exception("Error when opening Camera: " + str(err))

        self.vid_format = vid_format

        # Get in prepared mode and then set the video format
        ImagingSource.safe_call(ImagingSource.sdk.IC_PrepareLive, 1, self.cam)
        if vid_format is not None:
            ImagingSource.safe_call(ImagingSource.sdk.IC_SetVideoFormat, 1, self.cam, tis.T(vid_format))

        # Acquire the description of the image.
        width = ctypes.c_long()
        height = ctypes.c_long()
        bpp = ctypes.c_int()
        COLORFORMAT = ctypes.c_int()

        ImagingSource.safe_call(ImagingSource.sdk.IC_GetImageDescription, 1, self.cam, width, height, bpp, COLORFORMAT)

        # Dividing by 3 since it seems like even with format Y800 which is monochrome, it still uses 24 bits per pixel.
        # TODO: fix this to improve read efficiency
        bitdepth = int(bpp.value / 3)

        # Finally, use the superclass constructor to initialize other required variables.
        super().__init__(
            (width.value, height.value),
            bitdepth=bitdepth,
            name=serial,
            pitch_um=pitch_um,
            **kwargs
        )
        if verbose: print("success")

    def close(self):
        """See :meth:`.Camera.close`."""
        ImagingSource.safe_call(ImagingSource.sdk.IC_ReleaseGrabber, self.cam)
        del self.cam

    @staticmethod
    def info(verbose=True):
        """
        Discovers all cameras detected by the SDK.
        Useful for a user to identify the correct serial numbers / etc.

        Parameters
        ----------
        verbose : bool
            Whether to print the discovered information.

        Returns
        --------
        list of str
            List of serial numbers or identifiers.
        """
        if tis is None:
            raise ImportError("tisgrabber not installed. Install to use ImagingSource cameras.")

        if ImagingSource.sdk is None:
            err = ImagingSource.init_sdk()
            if err != 1:
                raise Exception("Error when loading SDK: " + str(err))

        # Get device count and then iterate through each device
        devicecount = ImagingSource.sdk.IC_GetDeviceCount()
        serial_list = []
        for i in range(0, devicecount):
            serial_list.append(tis.D(ImagingSource.sdk.IC_GetUniqueNamefromList(i)))

        if verbose: print(serial_list)

        return serial_list

    ### Property Configuration ###

    def _get_exposure_hw(self):
        """See :meth:`.Camera._get_exposure_hw`."""
        exposure = ctypes.c_float()
        ImagingSource.safe_call(ImagingSource.sdk.IC_GetPropertyAbsoluteValue, 1, self.cam, tis.T("Exposure"), tis.T("Value"), exposure)
        return float(exposure.value)

    def _set_exposure_hw(self, exposure_s):
        """See :meth:`.Camera._set_exposure_hw`."""
        # Turn off auto exposure and use the value given.
        ImagingSource.safe_call(ImagingSource.sdk.IC_SetPropertySwitch, 1, self.cam, tis.T("Exposure"), tis.T("Auto"), 0)
        ImagingSource.safe_call(ImagingSource.sdk.IC_SetPropertyAbsoluteValue, 1, self.cam, tis.T("Exposure"), tis.T("Value"), ctypes.c_float(exposure_s))

    def set_woi(self, woi=None):
        """See :meth:`.Camera.set_woi`."""
        if woi is None:
            width = ctypes.c_long()
            height = ctypes.c_long()
            bpp = ctypes.c_int()    # Bits per pixel
            COLORFORMAT = ctypes.c_int()

            ImagingSource.safe_call(ImagingSource.sdk.IC_GetImageDescription, 1, self.cam, width, height, bpp, COLORFORMAT)

            width = width.value
            height = height.value
            self.woi = (0, width, 0, height)
        else:
            # Format for woi is specified in :meth:`.Camera.set_woi`.
            width = int(woi[1])
            height = int(woi[3])
            xpos = int(woi[0])
            ypos = int(woi[2])
            # This keeps the original format
            idx = self.vid_format.find("(")    # TODO: is this general?
            this_vid_format = self.vid_format[:idx]
            # We bring in the new width and height specified in the video format
            tot_format = this_vid_format + "(" + str(width) +"x" + str(height) + ")"
            ImagingSource.safe_call(ImagingSource.sdk.IC_SetVideoFormat, 1, self.cam, tis.T(tot_format))
            # Now offset
            ImagingSource.safe_call(ImagingSource.sdk.IC_SetPropertySwitch, 1, self.cam, tis.T("Partial scan"), tis.T("Auto-center"), 0)
            ImagingSource.safe_call(ImagingSource.sdk.IC_SetPropertyValue, 1, self.cam, tis.T("Partial scan"), tis.T("X Offset"), xpos)
            ImagingSource.safe_call(ImagingSource.sdk.IC_SetPropertyValue, 1, self.cam, tis.T("Partial scan"), tis.T("Y Offset"), ypos)

        self.shape = (height, width)

    def _get_image_hw(self, timeout_s):
        """See :meth:`.Camera.get_image`."""
        buffer_size = 3 * self.bitdepth * self.shape[0] * self.shape[1] # times 3 is because even Y800 is RGB
        # Starts the image acquisition
        ImagingSource.safe_call(ImagingSource.sdk.IC_StartLive, 0, self.cam, 0)
        # Snap image
        err = ImagingSource.safe_call(ImagingSource.sdk.IC_SnapImage, 0, self.cam, 1000*timeout_s)
        # If there is an error, then snap image again
        while err <= 0:
            err = ImagingSource.safe_call(ImagingSource.sdk.IC_SnapImage, 0, self.cam, 1000*timeout_s)
        # Get image
        ptr = ImagingSource.safe_call(ImagingSource.sdk.IC_GetImagePtr, 0, self.cam)
        img_ptr = ctypes.cast(ptr, ctypes.POINTER(ctypes.c_ubyte * buffer_size))
        # Reshape the image according to the width and height.
        # TODO: there are more efficient ways to reshape the array only considering the R component.
        img = np.ndarray(buffer=img_ptr.contents, dtype=np.uint8, shape=(self.shape[0], self.shape[1], 3)) # 3 for RGB
        ImagingSource.safe_call(ImagingSource.sdk.IC_StopLive, 0, self.cam)
        # We take only the 1st component, assuming that the image is monochromatic.
        return self.transform(img[:,:,0])


"""
Hardware control for Santec SLMs.
Tested with Santec LCoS SLM-200, SLM-210, and SLM-300.

Note
~~~~
:class:`.Santec` requires dynamically linked libraries from Santec to be present in the
runtime directory:

- SLMFunc.dll
- FTD3XX.dll

These files should be copied in before use.

Note
~~~~
Santec provides base wavefront correction accounting for the curvature of the SLM surface.
Consider loading these files via :meth:`.SLM.load_vendor_phase_correction()`
"""
import os
import ctypes
import numpy as np
import cv2
import warnings

from .slm import SLM

try:  # Load Santec's header file.
    from . import _slm_win as slm_funcs
except BaseException as e:  # Provide an informative error should something go wrong.
    warnings.warn(
        "Santec DLLs not installed. Install these to use Santec SLMs."
        "  Dynamically linked libraries from Santec (usually provided via USB) "
        "must be present in the runtime directory:\n"
        "  - SLMFunc.dll\n  - FTD3XX.dll\n"
        "  Check that these files are present and are error-free.\n"
        "Original error: {}".format(e)
    )
    slm_funcs = None


class Santec(SLM):
    """
    Interfaces with Santec SLMs.

    Attributes
    ----------
    slm_number : int
        USB port number assigned by Santec SDK.
    display_number : int
        Display number assigned by Santec SDK.
    optionboard_id : str
        ID of the option board
    driveboard_id : str
        ID of the drive board
    product_code_id : str
        Product code of the device
    """

    def __init__(
            self,
            slm_number=1,
            display_number=2,
            bitdepth=10,
            wav_um=1,
            pitch_um=(8,8),
            verbose=True,
            **kwargs
        ):
        r"""
        Initializes an instance of a Santec SLM.

        Arguments
        ---------
        slm_number
            See :attr:`slm_number`.
        display_number
            See :attr:`display_number`.
        bitdepth : int
            Depth of SLM pixel well in bits. Defaults to 10.
        wav_um : float
            Wavelength of operation in microns. Defaults to 1 μm.
        pitch_um : (float, float)
            Pixel pitch in microns. Defaults to 8 micron square pixels.
        verbose : bool
            Whether to print extra information.
        **kwargs
            See :meth:`.SLM.__init__` for permissible options.

        Note
        ----
        Santec SLMs can reconfigure their phase table: the correspondence between
        grayscale values and applied voltages. This is configured based upon the wavelength
        supplied to :attr:`.SLM.wav_design_um`. This allows :attr:`.SLM.phase_scaling`
        to be one if desired, and make use of optimized routines (see :meth`.set_phase()`).
        However, sometimes setting the phase table runs into issues, where the maximum value
        doesn't correspond to exactly :math:`2\pi` at the target wavelength. This is noted
        in the initialization, and the user should update :attr:`.SLM.wav_design_um` or otherwise
        to avoid undesired behavior.
        :attr:`wav_design_um` defaults to :attr:`wav_um` if it is not given.

        Caution
        ~~~~~~~
        :class:`.Santec` defaults to 8 micron square pixel size :attr:`.SLM.pitch_um`
        and 10-bit :attr:`.SLM.bitdepth`.
        This is valid for SLM-200, SLM-210, and SLM-300,
        but may not be valid for future Santec models.
        """
        if slm_funcs is None:
            raise RuntimeError(
                "Santec DLLs not installed. Install these to use Santec SLMs."
                "  Dynamically linked libraries from Santec (usually provided via USB) "
                "must be present in the runtime directory:\n"
                "  - SLMFunc.dll\n  - FTD3XX.dll\n"
                "  Check that these files are present and are error-free."
            )

        # Default max phase. Maybe this should be opened to the user in the future.
        # Otherwise, wav_um and wav_design_um have the same functionality.
        max_phase = 2 * np.pi

        self.slm_number = int(slm_number)
        self.display_number = int(display_number)

        # By default, target wavelength is the design wavelength
        wav_design_um = kwargs.pop("wav_design_um", None)
        if wav_design_um is None:
            wav_design_um = wav_um

        if verbose:
            print("Santec slm_number={} initializing... ".format(self.slm_number), end="")
        Santec._parse_status(slm_funcs.SLM_Ctrl_Open(self.slm_number))

        try:
            # Wait for the SLM to no longer be busy.
            while True:
                status = slm_funcs.SLM_Ctrl_ReadSU(self.slm_number)

                if status == 0:
                    break  # SLM_OK (proceed)
                elif status == 2:
                    continue  # SLM_BS (busy)
                else:
                    Santec._parse_status(status)

            # Check to see if the device or option boards have an error.
            self.get_error(raise_error=True)

            # Right now, only DVI mode is supported.
            Santec._parse_status(slm_funcs.SLM_Ctrl_WriteVI(self.slm_number, 1))  # 0:Memory 1:DVI
            if verbose:
                print("success")

            # Update wavelength if needed
            wav_current_nm = ctypes.c_uint32(0)
            phase_current = ctypes.c_ulong(0)
            Santec._parse_status(
                slm_funcs.SLM_Ctrl_ReadWL(self.slm_number, wav_current_nm, phase_current)
            )

            wav_desired_nm = int(1e3 * wav_design_um)
            phase_desired = int(np.floor(max_phase * 100 / np.pi))

            # Update the phase table if necessary. This sometimes fails for unknown
            # reasons, so we do multiple attempts. Reasons for failure might include overheating
            # while the energy-intensive process of updating the table is underway.
            attempt = 1
            while wav_current_nm.value != wav_desired_nm and attempt < 5:
                if verbose:
                    if attempt > 1:
                        print("(attempt {})".format(attempt))
                    else:
                        print(
                            "Current phase table: wav = {0} nm, maxphase = {1:.2f}pi".format(
                                wav_current_nm.value, phase_current.value / 100.0
                            )
                        )
                        print(
                            "Desired phase table: wav = {0} nm, maxphase = {1:.2f}pi".format(
                                wav_desired_nm, phase_desired / 100.0
                            )
                        )
                    print("     ...Updating phase table (this may take 40 seconds)...")

                # Set wavelength (nm) and maximum phase (100 * [float pi])
                Santec._parse_status(
                    slm_funcs.SLM_Ctrl_WriteWL(
                        self.slm_number, ctypes.c_uint32(wav_desired_nm), phase_desired
                    )
                )

                # Save wavelength
                Santec._parse_status(slm_funcs.SLM_Ctrl_WriteAW(self.slm_number))

                # Verify wavelength
                Santec._parse_status(
                    slm_funcs.SLM_Ctrl_ReadWL(self.slm_number, wav_current_nm, phase_current)
                )
                if verbose:
                    print(
                        "Updated phase table: wav = {0} nm, maxphase = {1:.2f}pi".format(
                            wav_current_nm.value, phase_current.value / 100.0
                        )
                    )

                attempt += 1

            # Raise an error if we failed.
            if wav_current_nm.value != wav_desired_nm or abs(phase_current.value - 200) > 100:
                raise RuntimeError("Failed to update Santec phase table.")

            # Note phase table issues if they are present
            if verbose and abs(phase_current.value - 200) > 4:
                wav_design_fixed_um = wav_design_um * (phase_current.value / 200.0)
                print(
                    "  Warning: the Santec phase table maximum deviates significantly (>2%) from 2pi ({0:.2f}pi).".format(
                        phase_current.value / 100.0
                    )
                )
                print(
                    "    This is likely due to internal checks avoiding 'abnormal' phase table results."
                )
                print(
                    "    To compensate for this, wav_design_um is noted to equal {} instead of the desired {}.".format(
                        wav_design_fixed_um, wav_design_um
                    )
                )
                if wav_um / wav_design_fixed_um != 1:
                    print(
                        "    This results in phase_scaling={0:.4f} != 1, which has negative speed implications (see .set_phase()).".format(
                            wav_um / wav_design_fixed_um
                        )
                    )
                print(
                    "    If this behavior is undesired, play with wav_design_um to find a better regime."
                )
                wav_design_um = wav_design_fixed_um

            # Check for the SLM parameters and save them
            width = ctypes.c_ushort(0)
            height = ctypes.c_ushort(0)
            display_name = ctypes.create_string_buffer(128)

            if verbose:
                print("Looking for display_number={}... ".format(self.display_number), end="")
            Santec._parse_status(
                slm_funcs.SLM_Disp_Info2(self.display_number, width, height, display_name)
            )

            # For instance, "LCOS-SLM,SOC,8001,2018021001"
            # Format is "UserFriendlyName,ManufacterName,ProductCodeID,SerialNumberID"
            name = display_name.value.decode("mbcs")
            if verbose:
                print("success")

            names = name.split(",")

            # If the target display_number is not found, then print some debug:
            if names[0] != "LCOS-SLM":
                # Don't parse status around this one...
                raise ValueError(
                    "SLM not found at display_number={}. "
                    "Use .info() to find the correct display!".format(self.display_number)
                )

            # Populate some info
            driveboard_id = ctypes.create_string_buffer(16)
            optionboard_id = ctypes.create_string_buffer(16)
            Santec._parse_status(
                slm_funcs.SLM_Ctrl_ReadSDO(self.slm_number, driveboard_id, optionboard_id)
            )

            self.driveboard_id = driveboard_id.value.decode("mbcs")
            self.optionboard_id = optionboard_id.value.decode("mbcs")
            self.product_code_id = names[2]

            # Open SLM
            if verbose:
                print("Opening {}... ".format(name), end="")
            Santec._parse_status(slm_funcs.SLM_Disp_Open(self.display_number))
            if verbose:
                print("success")

            super().__init__(
                (int(width.value), int(height.value)),
                bitdepth=bitdepth,
                name=kwargs.pop("name", names[-1]), # SerialNumberID
                wav_um=wav_um,
                wav_design_um=wav_design_um,
                pitch_um=pitch_um,
                **kwargs,
            )

            self.set_phase(None)
        except Exception as init_error:
            try:
                Santec._parse_status(slm_funcs.SLM_Ctrl_Close(self.slm_number))
            except Exception as close_error:
                print(
                    "Could not close attempt to open Santec slm_number={}: {}".format(
                        slm_number, str(close_error)
                    )
                )

            raise init_error

    @staticmethod
    def info(verbose=True):
        """
        Discovers the names of all the displays.
        Checks all 8 possible supported by Santec's SDK.

        Parameters
        ----------
        verbose : bool
            Whether to print the discovered information.

        Returns
        --------
        list of (int, str) tuples
            The number and name of each potential display.
        """
        if slm_funcs is None:
            raise RuntimeError(
                "Santec DLLs not installed. Install these to use Santec SLMs."
                "  Dynamically linked libraries from Santec (usually provided via USB) "
                "must be present in the runtime directory:\n"
                "  - SLMFunc.dll\n  - FTD3XX.dll\n"
                "  Check that these files are present and are error-free."
            )

        # Check for the SLM parameters and save them
        display_list = []

        if verbose:
            print("Displays detected by Santec")
            print("display_number, display_name:")

        for display_number in range(1, 9):
            width = ctypes.c_ushort(0)
            height = ctypes.c_ushort(0)
            display_name = ctypes.create_string_buffer(128)

            status = slm_funcs.SLM_Disp_Info2(display_number, width, height, display_name)

            if not (status == 0 or status == -1):
                Santec._parse_status(status)

            name = display_name.value.decode("mbcs")
            if len(name) > 0:
                if verbose:
                    print("{},  {}".format(display_number, name))

                display_list.append((display_number, name))

        return display_list

    def load_vendor_phase_correction(self, file_path, smooth=False, overwrite=True):
        """
        Load phase correction provided by Santec from file,
        setting ``"phase"`` in :attr:`~slmsuite.hardware.slms.slm.SLM.source`.

        Parameters
        ----------
        file_path : str
            File path for the vendor-provided phase correction.
        smooth : bool
            Whether to apply a Gaussian blur to smooth the data.
        overwrite : bool
            Whether to overwrite the previous phase in
            :attr:`~slmsuite.hardware.slms.slm.SLM.source`.

        Note
        ~~~~
        This correction is only fully valid at the wavelength at which it was collected.

        Returns
        ----------
        numpy.ndarray
            :attr:`~slmsuite.hardware.slms.slm.SLM.source` ``["phase"]``,
            the Santec-provided phase correction.
        """
        try:
            # Load from .csv, skipping the first row and column
            # (corresponding to X and Y coordinates).
            map = np.loadtxt(file_path, skiprows=1, dtype=int, delimiter=",")[:, 1:]
            phase = (-2 * np.pi / self.bitresolution) * map.astype(float)

            # Smooth the map
            if smooth:
                real = np.cos(phase)
                imag = np.sin(phase)
                size_blur = 15  # The user should have access to this eventually

                real = cv2.GaussianBlur(real, (size_blur, size_blur), 0)
                imag = cv2.GaussianBlur(imag, (size_blur, size_blur), 0)

                # Recombine the components
                phase = np.arctan2(imag, real) + np.pi

            if overwrite:
                self.source["phase"] = phase

            return phase
        except BaseException as e:
            warnings.warn("Error while loading phase correction.\n{}".format(e))
            return self.source["phase"]

    def close(self):
        """See :meth:`.SLM.close`."""
        slm_funcs.SLM_Disp_Close(self.display_number)
        slm_funcs.SLM_Ctrl_Close(self.slm_number)

    def _set_phase_hw(self, display):
        """
        Hardware-specific implementation for Santec SLM devices.

        See :meth:`SLM._set_phase_hw` for the base class documentation.

        Parameters
        ----------
        display
            Integer data to display on the SLM. See :meth:`.SLM._set_phase_hw`.
        """
        matrix = display.astype(slm_funcs.USHORT)
        n_h, n_w = self.shape

        # Write to SLM
        c = matrix.ctypes.data_as(ctypes.POINTER((slm_funcs.USHORT * n_h) * n_w)).contents
        Santec._parse_status(
            slm_funcs.SLM_Disp_Data(self.display_number, n_w, n_h, 0, c), raise_error=False
        )

    ### Additional Santec-specific functionality

    def get_temperature(self):
        """
        Read the drive board and option board temperatures.

        Returns
        -------
        (float, float)
            Temperature in Celsius of the drive and option board
        """
        # Note that the Santec documentation suggests using signed
        # integers, but the header requests unsigned integers.
        drive_temp = ctypes.c_uint32(0)
        option_temp = ctypes.c_uint32(0)

        Santec._parse_status(slm_funcs.SLM_Ctrl_ReadT(self.slm_number, drive_temp, option_temp))

        return (drive_temp.value / 10.0, option_temp.value / 10.0)

    def get_error(self, raise_error=True, return_codes=False):
        """
        Read the drive board and option board errors.

        Parameters
        ----------
        raise_error : bool
            Whether to raise an error (if True) or a warning (if False) if error(s) are detected.
        return_codes : bool
            Whether to return an error string or integer error codes
            (in ``(drive_error, option_error)`` form).

        Returns
        -------
        list of str OR (int, int)
            List of errors.
        """
        drive_error = ctypes.c_uint32(0)
        option_error = ctypes.c_uint32(0)

        Santec._parse_status(
            slm_funcs.SLM_Ctrl_ReadEDO(self.slm_number, drive_error, option_error),
            raise_error=raise_error,
        )

        # Check the resulting bitstrings for errors (0 ==> all good).
        errors = []

        for drive_error_bit in slm_funcs.SLM_DRIVEBOARD_ERROR.keys():
            if drive_error.value & drive_error_bit:
                errors.append(slm_funcs.SLM_DRIVEBOARD_ERROR[drive_error_bit])

        for option_error_bit in slm_funcs.SLM_OPTIONBOARD_ERROR.keys():
            if option_error.value & option_error_bit:
                errors.append(slm_funcs.SLM_OPTIONBOARD_ERROR[option_error_bit])

        if raise_error and len(errors) > 0:
            error = "Santec error: " + ", ".join(["'" + err + "'" for err in errors])
            if raise_error:
                raise RuntimeError(error)
            else:
                warnings.warn(error)

        if return_codes:
            return (drive_error.value, option_error.value)
        else:
            return errors

    def get_status(self, raise_error=True):
        """
        Gets ``SLM_STATUS`` return from a Santec SLM and parses the result.

        Parameters
        ----------
        raise_error : bool
            Whether to raise an error (if True) or a warning (if False) when status is
            not ``SLM_OK``.

        Returns
        -------
        (int, str, str)
            Status in ``(num, name, note)`` form.
        """
        return Santec._parse_status(slm_funcs.SLM_Ctrl_ReadSU(self.slm_number), raise_error)

    @staticmethod
    def _parse_status(status, raise_error=True):
        """
        Parses the meaning of a ``SLM_STATUS`` return from a Santec SLM.

        Parameters
        ----------
        status : int
            ``SLM_STATUS`` return.
        raise_error : bool
            Whether to raise an error (if True) or a warning (if False) when status is not ``SLM_OK``.

        Returns
        -------
        (int, str, str)
            Status in ``(name, note)`` form.
        """
        # Parse status
        status = int(status)

        if not status in slm_funcs.SLM_STATUS_DICT.keys():
            raise ValueError("SLM status '{}' not recognized.".format(status))

        # Recover the meaning of status
        (name, note) = slm_funcs.SLM_STATUS_DICT[status]

        status_str = "Santec error {}; '{}'".format(name, note)

        if status != 0:
            if raise_error:
                raise RuntimeError(status_str)
            else:
                warnings.warn(status_str)

        return (status, name, note)

    def load_csv(self, filename):
        """
        Write the phase image contained in a .csv file to the SLM.
        This image should have the size of the SLM.

        Parameters
        ----------
        filename : str
            Path to the .csv file.
        """
        Santec._parse_status(slm_funcs.SLM_Disp_ReadCSV(self.display_number, 0, filename))

    ### Low priority trigger and memory features.
    ### Feel free to finish, or poke regarding implementation.

    # def software_trigger(self):
    #     slm_funcs.SLM_Ctrl_WriteTS(self.slm_number)

    # def set_intput_trigger(self, enabled=True):
    #     slm_funcs.SLM_Ctrl_WriteTI(self.slm_number, int(enabled))

    # def set_intput_trigger_direction(self, ascending=True):
    #     slm_funcs.SLM_Ctrl_WriteTC(self.slm_number, int(ascending))

    # def set_output_trigger(self, enabled=True):
    #     slm_funcs.SLM_Ctrl_WriteTM(self.slm_number, int(enabled))

    # def set_memory_framerate(self, framerate_hz):
    #     # Not sure if it's actually Hz
    #     assert framerate_hz >= 0
    #     assert framerate_hz <= 120
    #     slm_funcs.SLM_Ctrl_WriteMW(self.slm_number, int(framerate_hz))

    # def get_memory_framerate(self):
    #     framerate_hz = slm_funcs.DWORD(0)

    #     slm_funcs.SLM_Ctrl_ReadMW(self.slm_number, framerate_hz)

    #     return int(framerate_hz.value)

    # def start_memory_continuous(self, ascending=True):
    #     slm_funcs.SLM_Ctrl_WriteDR(self.slm_number, int(ascending))

    # def stop_memory_continuous(self):
    #     slm_funcs.SLM_Ctrl_WriteDB(self.slm_number)

    # def set_memory_table(self, mapping):
    #     assert len(mapping) == 128

    #     for entry, map in zip(range(1, 129), mapping):
    #         Santec._parse_status(
    #             slm_funcs.SLM_Ctrl_WriteMT(self.slm_number, entry, int(map))
    #         )

    # def get_memory_table(self):
    #     table = []

    #     ptr = slm_funcs.DWORD(0)

    #     for entry in range(1, 129):
    #         Santec._parse_status(
    #             slm_funcs.SLM_Ctrl_ReadMS(self.slm_number, entry, ptr)
    #         )
    #         table.append(int(ptr.value))

    #     return table

    # def set_memory_table_position(self, position):
    #     slm_funcs.SLM_Ctrl_WriteMP(self.slm_number, int(position))

    # def get_memory_table_position(self, position):
    #     # Function does not exist?
    #     slm_funcs.SLM_Ctrl_ReadMP(self.slm_number, int(position))

    # def set_memory_table_range(self, start, end):
    #     # end is one above the true end, according to docs.
    #     slm_funcs.SLM_Ctrl_WriteMR(self.slm_number, int(start), int(end))

    # def get_memory_table_range(self):
    #     start = slm_funcs.DWORD(0)
    #     end = slm_funcs.DWORD(0)

    #     slm_funcs.SLM_Ctrl_ReadMR(self.slm_number, start, end)

    #     return (start.value, end.value)

    # def set_phase_memory(self, phase):
    #     # Needs to be combined with set_phase(), probably.
    #     slm_funcs.SLM_Ctrl_WriteMI()
    #     raise NotImplementedError()
